# import re
# import numpy as np
# import pkg_resources
# import yaml
# import inspect
#
# import gherkan.utils.constants as c
#
# class SignalCondition():
#     def __init__(self, string, kind, negated=False):
#         self.kind = kind
#         self.string = string
#         self.variable = None
#         self.value = None
#         self.variableTr = None
#         self.templateID = None
#         self.variableNLP = None
#         self.variableNLPCZ = None
#         self.variableNL_full = None
#         self.variableNL_full_cs = None
#
#         if self.kind == c.EQUALITY or kind == c.INEQUALITY:
#             match = re.search(r'(?P<variable>\w+)\s*[=!]+\s*(?P<value>\w+)', self.string)
#             if negated:
#                 if self.kind == c.EQUALITY:
#                     self.kind = c.INEQUALITY
#                 else:
#                     self.kind = c.EQUALITY
#         elif self.kind == c.BOOL:
#             match = re.search(r'(?P<variable>\w+)', self.string)
#             self.value = 1 if not negated else 0
#         elif self.kind == c.EDGE:
#             match = re.search(r'edge\((?P<variable>\w+)\s*,\s*(?P<value>\w+)', self.string)
#         elif self.kind == c.FORCE:
#             match = re.search(r'force\((?P<variable>\w+)\s*,\s*(?P<value>\w+)', self.string)
#         elif self.kind == c.UNFORCE:
#             match = re.search(r'unforce\((?P<variable>\w+)', self.string)
#         else:
#             raise Exception("Error while parsing a signal condition (whaaat?). Related string: '{}'. Unexpected condition kind: '{}'.".format(self.string, self.kind))
#         if match:
#             self.variable = match.group("variable")
#             if "value" in match.groupdict().keys():
#                 self.value = match.group("value")
#         else:
#             raise Exception("Error while parsing a signal condition. Phrase matching failed!\nRelated string: '{}'\nCondition kind: '{}'".format(self.string, self.kind))
#
#
# class NLCondition():
#     def __init__(self, strFullEn = None, strFullCs = None, strEng = None, strCZ = None):
#         self.kind = None
#         self.string = strFullEn
#         self.variable = None
#         self.value = None
#         self.variableTr = None
#         self.templateID = None
#         self.variableNL_full = strFullEn
#         self.variableNL_full_cs = strFullCs
#         self.variableNLP = strEng
#         self.variableNLPCZ = strCZ
#
#
# # TODO defined elsewhere, remove
# # class SignalStatementTreeNode():
# #
# #     def __init__(self, kind=None, data=None):
# #         self.__kind = kind
# #         self.__data = data
# #         self.__parent = None
# #
# #     def setParent(self, parent):
# #         self.__parent = parent
# #
# #     @property
# #     def kind(self):
# #         return self.__kind
# #
# #     @property
# #     def parent(self):
# #         return self.__parent
# #
# #     @property
# #     def data(self):
# #         return self.__data
# #
# #     def getIncompleteParent(self):
# #         raise NotImplementedError()
# #
# #
# #
# # class SignalStatementTreeOperatorNode(SignalStatementTreeNode):
# #     AND = 'and'
# #     OR = 'or'
# #
# #     def __init__(self, kind):
# #         super().__init__(kind.lower())
# #         self.__lchild = None  # type: SignalStatementTreeNode
# #         self.__rchild = None  # type: SignalStatementTreeNode
# #         self.templateID = None
# #         self.variableNLP = None
# #         self.variableNL_full = None
# #         self.variableNLPCZ = None
# #         self.variableNL_full_cs = None
# #
# #     @property
# #     def lchild(self):
# #         return self.__lchild
# #
# #     @property
# #     def rchild(self):
# #         return self.__rchild
# #
# #     @property
# #     def children(self):
# #         return [self.lchild, self.rchild]
# #
# #     @property
# #     def full(self):
# #         return (self.lchild is not None) and (self.rchild is not None)
# #
# #     def getIncompleteParent(self):
# #         if self.full:
# #             if self.parent:
# #                 return self.parent.getIncompleteParent()
# #             else:
# #                 return None
# #         else:
# #             return self
# #
# #     def attachChild(self, child: SignalStatementTreeNode):
# #         """
# #         Returns True if the node is "full" (i.e. both children are assigned)
# #         """
# #         child.setParent(self)
# #         if self.lchild:
# #             self.__rchild = child
# #             return True
# #         else:
# #             self.__lchild = child
# #             return False
# #
# #     def __str__(self):
# #         return 'Operator(type: {})'.format(self.kind.upper())
#
#
# # class SignalStatementTreeOperandNode(SignalStatementTreeNode):
# #
# #     def __init__(self, data: SignalCondition):
# #         super().__init__(data.kind, data)
# #
# #     def getIncompleteParent(self):
# #         self.parent.getIncompleteParent()
# #
# #     def __str__(self):
# #         return 'Operand(type: {}, variable: {}, value: {}, variableNLP: {}, variableNL_full: {})'.format(self.kind.upper() if self.kind else 'None', self.data.variable, self.data.value, self.data.variableNLP, self.data.variableNL_full)
# #
# #     @property
# #     def variable(self):
# #         return self.data.variable
# #
# #     @property
# #     def kind(self):
# #         return self.data.kind
# #
# #     @property
# #     def variableNLP(self):
# #         return self.data.variableNLP
# #
# #     @property
# #     def variableNL_full(self):
# #         return self.data.variableNL_full
# #
# #     @property
# #     def value(self):
# #         return self.data.value
# #
# #
# # class SignalStatementTree():
# #     dash, line, left, right = '\u2500', '\u2502', '\u251c\u2500\u2500 ', '\u2514\u2500\u2500 '
# #
# #     def __init__(self, root: SignalStatementTreeNode, string: str):
# #         self.__root = root
# #         self.__string = string
# #
# #     @property
# #     def root(self):
# #         return self.__root
# #
# #     @property
# #     def string(self):
# #         return self.__string
# #
# #     def __printNode(self, node: SignalStatementTreeNode, leftChild, levelsOpened):
# #         string = ''.join([SignalStatementTree.line + "\t" if lo else "\t" for lo in levelsOpened[:-1]])
# #         if leftChild:
# #             if node:
# #                 string += SignalStatementTree.left + str(str(node) if bool(node) else "ERROR!!!") + '\n'
# #         else:
# #             if node:
# #                 string += SignalStatementTree.right + str(str(node) if bool(node) else "ERROR!!!") + '\n'
# #
# #         if type(node) is SignalStatementTreeOperatorNode:
# #             string += self.__printNode(node.lchild, True, levelsOpened + [True])
# #             string += self.__printNode(node.rchild, False, levelsOpened + [False])
# #         return string
# #
# #     def __str__(self):
# #         levelsOpened = [False]
# #
# #         node = self.root
# #         string = str(node) + '\n'
# #         if type(node) is SignalStatementTreeOperatorNode:
# #             string += self.__printNode(node.lchild, True, levelsOpened + [True])
# #             string += self.__printNode(node.rchild, False, levelsOpened + [False])
# #         return string
# #
#
# class EvalFunction():
#     """
#     Place holder for functions to be evaluated on the SignalStatementTree.
#     Contains additional functionality to enable creation of more complex templates.
#
#     Attributes
#     ----------
#     func : Function
#         The function to be evaluated on SignalStatementTreeNode objects (altough the implementation does not enforce any class)
#
#     succeeded : bool
#         True if the function was evaluated to True sufficient number of times
#
#     successes : int
#         The number of times when the function was evaluated successfully
#     """
#
#     def __init__(self, func, minSuccesses=1, maxSuccesses=None, failFast=False):
#         """
#         Parameters
#         ----------
#         func : Function
#             The function to be evaluated on SignalStatementTreeNode objects
#
#         minSuccesses : int
#             The minimum number of times the function needs to evaluate to True
#             to declare overall success. Use this in combination with maxSuccesses
#             to have a range of possible nodes that comply to the requirements
#             enforced by the function.
#
#         maxSuccesses : int
#             The maximum number of cases the function can be evaluated to True.
#             After the maxSuccesses has been reached, all following evaluations
#             will fail, even if they would succeed otherwise. Set this value
#             to a very large number to achieve unlimited number of successful
#             evaluations.
#
#         failFast : bool
#             If True, the first case when the function is evaluated to False
#             will result in complete failure of this condition. I.e. all future
#             evaluations will only fail and overall success will be False
#             no matter what. Use this to enforce a condition on all tested nodes.
#             failFast will also cause permanent failure if number of maxSuccesses
#             is exceeded. This can be used to limit the number of nodes in a subtree.
#             E.g., EvalFunction(lambda x: True, maxSuccesses=5, failFast=True)
#             Will generate a function that will make sure that there are at most 5
#             nodes in the subtree.
#         """
#         self.__func = func
#         self.__succeeded = None
#         self.__successes = 0
#         self.__minSuccesses = minSuccesses
#         self.__failFast = failFast
#         self.__valid = True
#         if self.__failFast:
#             self.__maxSuccesses = maxSuccesses if maxSuccesses else np.iinfo(np.int).max
#         else:
#             self.__maxSuccesses = maxSuccesses if maxSuccesses else minSuccesses
#
#     def invalidate(self):
#         self.__valid = False
#
#     def __bool__(self):
#         if self.__succeeded is None:
#             if self.successes >= self.minSuccesses:
#                 self.__succeeded = True
#                 return True
#             else:
#                 return False
#         elif self.__succeeded is True:
#             return True
#         elif self.__succeeded is False:
#             return False
#
#     def __call__(self, value):
#         """Calls the associated function and evaluates the result depending on other conditions as well.
#         If maxSuccesses has been exceeded, the call will fail no matter the actual result
#         (the function won't even be evaluated). It the number of successes (evaluations/calls resulting in True)
#         is smaller than maxSuccesses, the function, associated with this placeholder, is called with the provided value.
#         If the call returns True, number of successes is increased. The success, i.e. the result of the call is returned.
#         If failFast is set to True, any call that will result in return value equal to False will also fail the success
#         of the entire placeholder (EvalFunction instance)
#
#         Parameters
#         ----------
#         value : any
#             Function parameters
#
#         Returns
#         -------
#         bool
#             The success of the call (depending on other parameters)
#         """
#         if self.__succeeded is not None and not self.__succeeded:
#             return False
#         if self.successes >= self.maxSuccesses:
#             if self.failFast and self.func(value):
#                 self.__succeeded = False
#             return False
#         success = self.func(value)
#         if success:
#             self.__successes += 1
#         elif self.failFast:
#             self.__succeeded = False
#         return success
#
#     @property
#     def successes(self):
#         return self.__successes
#
#     @property
#     def succeeded(self):
#         return bool(self)
#
#     @property
#     def func(self):
#         return self.__func
#
#     @property
#     def minSuccesses(self):
#         return self.__minSuccesses
#
#     @property
#     def maxSuccesses(self):
#         return self.__maxSuccesses
#
#     @property
#     def valid(self):
#         return self.__valid
#
#     @property
#     def failFast(self):
#         return self.__failFast
#
#     def __repr__(self):
#         sig = inspect.signature(self.func)
#
#         def displayParam(p, d):
#             return "\t\t\t{}:\n{}".format(p, "\n".join(["\t\t\t\t{}".format(item) for item in d])) if type(d) is list else "\t\t\t{}: {}".format(p, d)
#
#         return "\n".join([
#             ">>> EvalFunction object <<<",
#             "\tNumber of successes = {}".format(self.successes),
#             "\tMinimum number of successes = {}".format(self.minSuccesses),
#             "\tMaximum number of successes = {}".format(self.maxSuccesses),
#             "\tSucceeded = {}".format(self.succeeded),
#             "\tAssociated function:",
#             "\t\tSignature: {}".format(str(sig)),
#             "\t\tParameters: ",
#             "\n".join(["\t\t\t" + param if sig.parameters[param].default is inspect._empty else displayParam(param, sig.parameters[param].default) for param in sig.parameters]),
#             "\t\tSource code: \n" + inspect.getsource(self.func).strip()
#         ])
#
#
# class EvalFunctionList():
#
#         def __init__(self, functionList):
#             """
#             Helper function to generate so called structure list from a list of functions.
#             Changed to generate a list of EvalFunction objects
#             For example:
#             [lambda x: x + 2, lambda x: x * 3] -> [EvalFunction(lambda x: x + 2), EvalFunction(lambda x: x * 3)]
#             """
#             self.evalList = []
#             for item in functionList:
#                 if callable(item):  # just the function; default arguments will be used
#                     evalList.append(EvalFunction(item))
#                 elif type(item) is tuple:  # a list of positional arguments
#                     evalList.append((EvalFunction(*item)))
#                 elif type(item) is dict:  # dictionary with kwargs
#                     evalList.append((EvalFunction(**item)))
#                 else:
#                     raise Exception("Unknown object type encountered while constructing structure list.")
#
# # TODO make iterable and validable
# # TODO incorporate additional methods into SignalStatementTree
#
# # class SignalTemplateTree(SignalStatementTree):
# #
# #     def __init__(self, signalTree: SignalStatementTree):
# #         super().__init__(signalTree.root, signalTree.string)
# #
# #     def __isTwig(self, node: SignalStatementTreeNode):
# #         """
# #         Checks if a node is an operator node but has only leaves (i.e. no consecutive operator nodes == no further branching)
# #         """
# #         if type(node) is SignalStatementTreeOperatorNode:
# #             return (type(node.lchild) is SignalStatementTreeOperandNode) and (type(node.rchild) is SignalStatementTreeOperandNode)
# #         else:
# #             return False
# #
# #     def __recursiveBoolEval(self, node: SignalStatementTreeNode, func):
# #         """
# #         Recursively attempts to evaluate the function on all subsequent nodes.
# #         Returns True if all evaluations were successful. That is, if the function
# #         is holds true for all nodes in the subtree.
# #         """
# #         result = func(node)
# #         if type(node) is SignalStatementTreeOperandNode:
# #             return result
# #         elif type(node) is SignalStatementTreeOperatorNode:
# #             return result and self.__recursiveBoolEval(node.lchild, func) and self.__recursiveBoolEval(node.rchild, func)
# #         else:
# #             # Unknown node type
# #             return False
# #
# #     def __recursiveStructureEval(self, node: SignalStatementTreeNode, structureList: list):
# #         """
# #         Recursively enters nodes and attempts to evalute functions in the structureList list.
# #         If one of the evaluations is successful it then attempts to continue with evaluations
# #         on the node's children.
# #
# #         Old version:
# #         "However, the successfully evaluated function is omitted
# #         from further evaluations. That is, each function must succeed exactly once!"
# #
# #         Newer version:
# #         "The number of successful evaluations depends on the specific conditions
# #         imposed by the EvalFunction object"
# #
# #         Functions are evaluated in a greedy manner. The order of the functions
# #         should therefore be considered - e.g., if some function A can succeed on more nodes
# #         than other function B in the list (i.e., domain of B(x) = True is a subset of A(x) = True),
# #         B should precede A in the list. Otherwise function A might take node that should
# #         normally be captured by B.
# #         """
# #         if len(structureList) == 0:  # either empty list has been provided or some node emptied it (see !success condition below)
# #             return False
# #         success = False
# #         allTrue = True
# #         for i, evalFunc in enumerate(structureList):
# #             if not evalFunc.valid:
# #                 # TODO how to handle allTrue?????
# #                 continue
# #             # Run the eval function on the node.
# #             # The boolean value of the function object is assigned to the allTrue variable to check if the structure is completed
# #             # the internal success of the function is handled by the EvalFunction object
# #             if not success:  # if no function succeeded so far
# #                 # call the function with the current node as a parameter
# #                 success = evalFunc(node)  # if the call returned True, i.e. the node comforms to the condition imposed by the function
# #                 # capture the first successful evaluation and do not call any other function
# #                 # the list has to be iterated through though to check if all function succeeded already
# #             allTrue &= bool(evalFunc)
# #             if success and not allTrue:
# #                 break
# #         if allTrue:  # structureList has been completed
# #             return True
# #         if success:
# #             if type(node) is SignalStatementTreeOperatorNode:
# #                 # call function on both children
# #                 self.__recursiveStructureEval(node.lchild, structureList)
# #                 self.__recursiveStructureEval(node.rchild, structureList)
# #                 if len(structureList) > 0:  # if the list hasn't been emptied yet - the structure check did not fail so far
# #                     return all([bool(e) for e in structureList])  # return the current completion status of the structure check
# #             return False
# #         else:
# #             # if none of the functions in the list succeeded (i.e. the node is not part of the structure), then the structure check failed
# #             # that is, this was not the structure we were looking for...
# #             for evalFunc in structureList:
# #                 evalFunc.invalidate()
# #             return False
# #
# #     def __recursiveSubTreeSearch(self, node: SignalStatementTreeNode, func):
# #         """
# #         Returns all subtrees (returns their respective roots in a list, to be more exact),
# #         for which the function evalutes to True.
# #
# #         Parameters
# #         ----------
# #         node : SignalStatementTreeNode
# #             A node from the tree on which the function should be evaluted
# #
# #         func: Function
# #             Function to be evaluated
# #
# #         Returns
# #         -------
# #         list OR bool:
# #             List of roots of subtrees for which the function was evaluted to True.
# #             Returns just boolean True if this node evaluted to True. Otherwise,
# #             if this node evaluted to False but its child evaluted to True,
# #             a list containing this child will be returned.
# #         """
# #         if func(node):
# #             return node
# #         else:
# #             if type(node) is not SignalStatementTreeOperatorNode:
# #                 return False
# #             else:
# #                 rootList = []
# #                 # Evaluate the function on children and iterate over the results
# #                 for res in [self.__recursiveSubTreeSearch(node.lchild, func), self.__recursiveSubTreeSearch(node.rchild, func)]:
# #                     if not res:
# #                         continue
# #                     if type(res) is list:
# #                         rootList += res
# #                     elif type(res) is SignalStatementTreeNode:
# #                         rootList += [res]
# #
# #                 return rootList
# #
# #     def __genNextTemplate(self):
# #         """
# #         Generates function that should search for templates in the tree.
# #
# #         Raises:
# #             Exception -- Raises an exception if the ID of the requested template
# #             does not exist. That is, function for the template is not specified, yet.
# #         """
# #         with pkg_resources.resource_stream("gherkan", '/'.join(["containers", "templates_dic.yaml"])) as f:
# #             templateDict = yaml.load(f)
# #
# #         for templateID, templateSignature in templateDict.items():
# #             if "SignalTemplate" in templateSignature:  # find signal template that has the property called "SignalTemplate"
# #                 signalTemplateSignature = templateSignature["SignalTemplate"]
# #                 template = None
# #                 # find what kind of template it has (currently only structure is supported)
# #                 if type(signalTemplateSignature) is list or "structure" in signalTemplateSignature:
# #                     if type(signalTemplateSignature) is list:
# #                         templateStructure = signalTemplateSignature
# #                     elif type(signalTemplateSignature) is dict:
# #                         templateStructure = signalTemplateSignature["structure"]
# #                     structureList = []
# #                     for item in templateStructure:
# #                         if type(item) is str:
# #                             structureList.append(lambda nd, ndkind=item: type(nd) is SignalStatementTreeOperatorNode and nd.kind == ndkind)
# #                         elif type(item) is dict:
# #                             lamlist = []
# #                             kind, paramdict = item.popitem()
# #                             for key, value in paramdict.items():
# #                                 lamlist.append(lambda nd, var=key, val=value:  val.lower() in getattr(nd.data, var).lower())
# #                             structureList.append(lambda nd, ndkind=kind, paramCheckList=lamlist: type(nd) is SignalStatementTreeOperandNode and nd.kind == ndkind and all([check(nd) for check in paramCheckList]))
# #                         else:
# #                             raise Exception("Unknown type in signalTemplateSignature.\n{}".format(signalTemplateSignature))
# #
# #                     def template(node: SignalStatementTreeNode, structure=self.__genStructureFromFunctionList(structureList)):
# #                         def searchFunction(node: SignalStatementTreeNode):
# #                             return self.__recursiveStructureEval(node, structure)
# #                         return self.__recursiveSubTreeSearch(node, searchFunction)
# #
# #                 elif "condition" in signalTemplateSignature:
# #                     pass
# #                 elif "script" in signalTemplateSignature:
# #                     pass
# #
# #                 yield((templateID, template))
# #
# #     def parseTemplates(self):
# #         for templateID, template in self.__genNextTemplate():
# #             if not template:
# #                 raise Exception("Error while generating template from SignalTemplateTree!")
# #             tempResult = template(self.root)
# #             if tempResult:
# #                 print(tempResult)
# #                 if type(tempResult) is not list:
# #                     tempList = [tempResult]
# #                 for subtreeRoot in tempList:
# #                     subtreeRoot.templateID = templateID
#



import gherkan.utils.gherkin_keywords as g

sectionDict = {  # Dictionary of the possible sectioning denominators; add any synonymes and translations to each list
    g.FEATURE: [g.FEATURE, g.FEATURE_CS],
    g.BACKGROUND: [g.BACKGROUND, g.BACKGROUND_CS],
    g.SCENARIO: [g.SCENARIO, g.SCENARIO_CS],
    g.GIVEN: [g.GIVEN, g.GIVEN_CS],
    g.WHEN: [g.WHEN, g.WHEN_CS],
    g.THEN: [g.THEN, g.THEN_CS],
    g.AND: [g.AND, g.AND_CS],
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from flask import Flask, jsonify, render_template, make_response, request
from flask_restful import Resource, Api, reqparse
import os
import sys
from gherkan.flask_api import resources
from gherkan.flask_api import API_FSA
import argparse


def main(cmd_line_args=None):
    # Parse command line arguments
    parser = argparse.ArgumentParser(description='Runs RESTful API for the Gherkan NL Instruction Processing system.')

    host_group = parser.add_mutually_exclusive_group()
    host_group.add_argument('host', nargs="?", help="The hostname to listen on. Set this to '0.0.0.0' to have the server available externally as well. Defaults to '127.0.0.1'", default="127.0.0.1")
    host_group.add_argument('-x', '--external', help="Sets the host to '0.0.0.0' (see the 'host' parameter). Can only be used if host is not specified", action='store_true')

    port_group = parser.add_mutually_exclusive_group()
    port_group.add_argument('port', nargs="?", help='The port of the webserver. Defaults to 5000.', default=5000, type=int)
    port_group.add_argument('-p', help='The port of the webserver. Defaults to 5000.', default=0, type=int)

    parser.add_argument('-d', '--debug', help='Enables debug mode. An interactive debugger will be shown for unhandled exceptions, and the server will be reloaded when code changes.', action='store_true')
    parser.add_argument('-s', '--set', nargs=2, help='Sets a value to the fsa_config.yaml. Anything else is ignored, i.e. the system is not run and the script will exit after setting the variable. Multiple variables can be set at once.', action='append')
    parser.add_argument('-f', '--flush', help='Will flush previous error.', action='store_true')
    parser.add_argument('-rf', '--regenerate_flush', help='Regenerate fsa_config and flush errors.', action='store_true')

    args = parser.parse_args(cmd_line_args)
    print(args)

    api_host = args.host
    if args.set is not None:
        # Iterate of the parameters, set them and exit
        for (key, value) in args.set:
            if value.isdecimal():
                value = int(value)
            elif value.lower() == 'true':
                value = True
            elif value.lower() == 'false':
                value = False
            API_FSA.setConfig(key, value)
        sys.exit(0)
    else:
        if args.regenerate_flush:
            API_FSA.regenerateConfig()
            API_FSA.setState(API_FSA.S_OFF)
            API_FSA.setConfig("flush", True)
        elif args.flush:
            API_FSA.setConfig("flush", True)  # set flush to true so that previous errors are ignored
        if args.p > 0:
            api_port = args.p
        else:
            api_port = args.port
        api_debug = args.debug
        if args.external:
            api_host = '0.0.0.0'

    # Initialize the FSA
    API_FSA.init_fsa(api_host, api_port)

    app = Flask(__name__)
    api = Api(app)

    # Append resources
    api.add_resource(resources.Actions, '/actions', endpoint='actions')
    api.add_resource(resources.Actions, '/actions/<language>', endpoint='actions_language')
    api.add_resource(resources.Audio, '/audio')
    api.add_resource(resources.NLScenario, '/nlscenario')
    api.add_resource(resources.NLText, '/nltext')
    api.add_resource(resources.SignalMap, '/signal_map')
    api.add_resource(resources.SignalMap, '/signal_map/<language>', endpoint='signal_map_language')
    api.add_resource(resources.Signals, '/signals')
    api.add_resource(resources.Handler, '/<string:action>', endpoint='handler')
    api.add_resource(resources.Handler, '/<string:action>/<string:param>', endpoint='handler_param')

    # Run the API App
    app.run(host=api_host, port=api_port, debug=api_debug)


if __name__ == '__main__':
    print("setting up Gherkan API via script")
    main()



import os

from gherkan.decoder.SignalParser import SignalParser
import gherkan.utils.constants as c

# Path to the text file with the signal batch
input_filename = os.path.join(c.DATA_DIR, 'input', 'test_R1NormalSequence_signals.feature')

# Load the file into a SignalFileLoader object
signalParser = SignalParser()
# Parse the entire file
signalBatch = signalParser.parseFile(input_filename)

# Print stuff related to the entire batch, read from the file
print("Showing signal batch from file {}".format(input_filename))
print("Batch name: {}".format(signalBatch.name))
print("Batch description: {}".format(signalBatch.desc))
print("Batch context condition: {}".format(signalBatch.context))
print("Used language: {}".format(signalBatch.language))

dir = os.path.join(c.DATA_DIR, 'output')

# parse the signals file
sp = SignalParser()
signalBatch = sp.parseFile(input_filename)

for scenario in signalBatch.scenarios:
    print("Scenario: {}\n".format(scenario.name))
    for statementListType in ["contextStatements", "givenStatements", "thenStatements", "whenStatements"]:
        statementList = getattr(scenario, statementListType)

        if statementList:
            print("\t{}\n".format(statementListType))

            for statement in statementList:
                print("\t{}".format(statement))

            print("\t------------")




import datetime
import importlib
import os
import re
import atexit

import yaml

from .raw_text_to_signal import process as nlToSignal
from .signal_to_nl import process as signalToNL
from .speech_to_text import process as speechToText

from flask import send_file

DEBUG_MODE = True

class API_FSA():
    """
    FSA class for tracking state from REST API. State is written into a file
    so it should be persistent even over long time periods.
    """

    # =====>>> PATH variables <<<===== #
    stateFilePath = 'state.fsa'  # file to store the current state (see S_* variables)
    logFilePath = 'log.txt'  # log file
    configFilePath = 'fsa_config.yaml'  # (editable) configuration file

    # This path does not need to be set to the package's root, probably
    rootFolder = importlib.util.find_spec("gherkan").submodule_search_locations[0]  # the package's root folder

    dataFolder = os.path.join(rootFolder, 'data')  # folder with received and processed files (inp/out files)
    audioFolder = os.path.join(dataFolder, 'audio')  # folder with audio files

    robotProgramsPath = os.path.join(rootFolder, 'utils', 'RobotPrograms_{}.json')  # template of the path to the robot programs

    language = 'en'

    # =====>>> STATE constants <<<===== #
    # State is stored as 2 bytes in the state.fsa binary file
    S_OFF = 0
    S_IDLE = 1
    S_RECEIVED_NL_TEXT = 2
    S_CORRECTING_NL_TEXT = 4
    S_FINISHED_PROCESSING_NL_TEXT = 8
    S_RECEIVED_SIGNAL = 16
    S_FINISHED_PROCESSING_SIGNAL = 32
    S_RECEIVED_AUDIO = 64
    S_CORRECTING_AUDIO = 128
    S_FINISHED_PROCESSING_AUDIO = 256
    S_ERROR = 1024

    # This is for debugging
    STATES = ["S_IDLE", "S_RECEIVED_NL_TEXT", "S_CORRECTING_NL_TEXT", "S_FINISHED_PROCESSING_NL_TEXT", "S_RECEIVED_SIGNAL", "S_FINISHED_PROCESSING_SIGNAL", "S_RECEIVED_AUDIO", "S_CORRECTING_AUDIO", "S_FINISHED_PROCESSING_AUDIO", "S_ERROR"]

    # =====>>> PARAMETER name constants <<<===== #
    P_LAST_SIGNAL_FILE = "lastSignalFile"
    P_LAST_NL_TEXT_FILE = "lastNLTextFile"
    P_LAST_AUDIO_FILE = "lastAudioFile"
    P_RESPONSE_MODE = "responseMode"  # response mode for the processed text, see parameter VALUES

    # =====>>> parameter VALUES <<<===== #
    PV_RESPONSE_FILE = "responseModeFile"  # sends back a file
    PV_RESPONSE_JSON = "responseModeJSON"  # sends back a JSON

    # =====>>> OTHER <<<===== #
    basenameExtractorRegex = re.compile(r"(?P<basename>.+?)(?P<rest>(_signals|\.)+.+)")

    # =====>>> Helper STATIC methods <<<===== #
    @staticmethod
    def toBytes(integer):
        # Maybe extend this with a loop but so far 1024 as the limit is ok
        return bytes([(0b11100000000 & integer) >> 8, 0b11111111 & integer])

    @staticmethod
    def fromBytes(byte):
        # Maybe extend this with a loop but so far 1024 as the limit is ok
        return (byte[0] << 8) | byte[1]

    @staticmethod
    def generateDatestring():
        return datetime.datetime.now().strftime('%Y_%b_%d-%H_%M_%S')

    # =====>>> Helper CLASS methods <<<===== #
    @classmethod
    def getState(cls):
        with open(cls.stateFilePath, 'rb') as statefile:
            state = cls.fromBytes(statefile.read())
        if DEBUG_MODE:
            print(f"D>>>\nRequested state: {state}")
            states = [cls.STATES[i] for i in range(10) if (2**i) & state]
            print(f"Current states:\n{states}")
        return state

    @classmethod
    def setState(cls, state):
        if DEBUG_MODE:
            print(f"D>>>\nSetting state to: {state}")
            states = [cls.STATES[i] for i in range(10) if (2**i) & state]
            print(f"Current states:\n{states}")
        if isinstance(state, int):
            state = cls.toBytes(state)
        with open(cls.stateFilePath, 'wb') as statefile:
            statefile.write(state)

    @classmethod
    def addState(cls, state):
        """ Adds a state to the FSA
        This method should be preferred instead of the "setState" method
        """
        currentState = cls.getState()
        if state > cls.S_IDLE:
            currentState &= ~cls.S_IDLE
        cls.setState(currentState | state)
        
    @classmethod
    def removeState(cls, *states):
        """ Removes a state from the FSA
        This method should be preferred instead of the "setState" method.
        Multiple states can be removed at once
        This methods, however, cannot be used to set the state to S_OFF.
        """
        nextState = cls.getState()
        for state in states:
            nextState &= ~state
        if nextState < cls.S_IDLE:
            nextState |= cls.S_IDLE
        cls.setState(nextState)
        
    @classmethod
    def writeLog(cls, message):
        with open(cls.logFilePath, 'a') as logFile:
            logFile.writeline(message)

    @classmethod
    def readLog(cls):
        with open(cls.logFilePath, 'rb') as logFile:
            logFile.seek(-2, os.SEEK_END)
            while logFile.read(1) != b'\n':
                logFile.seek(-2, os.SEEK_CUR)
            line = logFile.readline().decode()
        return line

    @classmethod
    def getConfig(cls, param=None):
        with open(cls.configFilePath, 'r') as cfg:
            config = yaml.load(cfg)
            if param is not None:
                try:
                    value = config[param]
                except:
                    value = None
        if DEBUG_MODE:
            print(f"D>>>\nRequested value of configuration parameter: {param} = {value}")
        return value

    @classmethod
    def setConfig(cls, param, value):
        with open(cls.configFilePath, 'r') as cfg:
            config = yaml.load(cfg)
        config[param] = value
        with open(cls.configFilePath, 'w') as cfg:
            yaml.dump(config, cfg)
        if DEBUG_MODE:
            print(f"D>>>\nSetting configuration parameter: {param} = {value}")
            print("Configuration yaml file:")
            print(config)

    @classmethod
    def regenerateConfig(cls):
        if DEBUG_MODE:
            print("D>>>\nRegenerating configuration file")
        with open(cls.configFilePath, 'w') as cfg:
            # TODO: add more defaults to config
            yaml.dump({"flush": False, "responseMode": "responseModeFile"}, cfg)

    # =====>>> API initialization <<<===== #
    @classmethod
    def init_fsa(cls, address, port):
        print("The RESTful API to the Gherkan NL Instruction Processing system is starting...")

        # Check if the root folder is set correctly
        if not os.path.isdir(cls.rootFolder):
            raise Exception("The root folder is set to a non-existent path!\nRoot folder: {}".format(cls.rootFolder))
        if not os.access(cls.rootFolder, os.W_OK):
            raise Exception("Cannot write to the root folder!\nRoot folder: {}".format(cls.rootFolder))
        if not os.access(cls.rootFolder, os.R_OK):
            raise Exception("Cannot read from the root folder!\nRoot folder: {}".format(cls.rootFolder))
        
        # Check the existence of data folders
        if not os.path.isdir(cls.audioFolder):
            os.makedirs(cls.audioFolder)
        # if not os.path.isdir(cls.nlTextFolder):
        #     os.makedirs(cls.nlTextFolder)
        # if not os.path.isdir(cls.signalsFolder):
        #     os.makedirs(cls.signalsFolder)

        if not os.path.isfile(cls.configFilePath):
            cls.regenerateConfig()

        print("Host address: {}\nPort number: {}".format(address, port))
        cls.setConfig("host", address)
        cls.setConfig("port", port)

        print("Variable checking done, checking state from previous run...")
        # Check for previous sessions
        if os.path.exists(cls.stateFilePath):
            previousState = cls.getState()
            # Check and handle error
            if previousState == cls.S_ERROR:
                if cls.getConfig('flush'):
                    print("Found error state from a previous run.\nError flushing is set to True - flushing the error and reseting the system.")
                    previousState = cls.S_OFF
                    cls.regenerateConfig()
                else:
                    message = cls.readLog()
                    raise Exception("Previous run of the system resulted in an error! Resolve the error and set the fsa_config\nLast log message: {}".format(message))
            if previousState != cls.S_OFF:
                print("Found non-OFF state from previous run, restoring the system with state {}".format(previousState))
                cls.setState(previousState)
            else:
                # Set state to ready/idle
                print("System ready, state set to IDLE.")
                cls.setState(cls.S_IDLE)
        else:
            print("System ready, state set to IDLE.")
            cls.setState(cls.S_IDLE)

        atexit.register(cls.finalize)

    @classmethod
    def finalize(cls):
        print("Shutting down FSA.")
        cls.setState(cls.S_OFF)

    # =====>>> RESTful API Call handling functions <<<===== #
    @classmethod
    def receiveSignal(cls, text):
        """This function handles the receiving of signal text.
        """

        state = cls.getState()
        if state & (cls.S_RECEIVED_AUDIO | cls.S_RECEIVED_NL_TEXT):
            raise Exception("Different data type already received!")

        if state & cls.S_RECEIVED_SIGNAL:
            signalFileMode = 'a'
            fileName = cls.getConfig(cls.P_LAST_SIGNAL_FILE)
        else:
            cls.addState(cls.S_RECEIVED_SIGNAL)
            signalFileMode = 'w'
            fileName = os.path.join(cls.dataFolder, '_'.join([cls.generateDatestring(), 'signals.feature']))
            cls.setConfig(cls.P_LAST_SIGNAL_FILE, fileName)

        with open(fileName, signalFileMode) as sigFile:
            sigFile.writelines(text)

        # Signal text can be iteratively written into the file
        # When all signals were written, the user can request NL file

    @classmethod
    def requestSignal(cls):
        """This function return a composed signal file derived from NL text, if it was provided
        """
        state = cls.getState()
        if state & cls.S_RECEIVED_SIGNAL:  # Signals were sent and requested back (e.g. to check the state of the file)
            return send_file(cls.getConfig(cls.P_LAST_SIGNAL_FILE))
        elif not (state & cls.S_FINISHED_PROCESSING_NL_TEXT):  # Check if NL text was sent and processed
            raise Exception("NL text was not provided, cannot return signal file!")

        signalFilePath = cls.getConfig(cls.P_LAST_SIGNAL_FILE)

        responseMode = cls.getConfig(cls.P_RESPONSE_MODE)
        if responseMode == cls.PV_RESPONSE_FILE:
            response = send_file(signalFilePath)
        elif responseMode == cls.PV_RESPONSE_JSON:
            with open(signalFilePath, "r") as signalFile:
                signalText = signalFile.read()

            # TODO: Split file into response

            response = {
                    "language": "en<OR>cs",
                    "background": "background text",
                    "description": "This feature is not implemented, yet! Use RESPONSE_MODE: FILE instead.",
                    "scenarios": signalText
            }

        cls.removeState(cls.S_FINISHED_PROCESSING_NL_TEXT, cls.S_RECEIVED_NL_TEXT)
        return response

    @classmethod
    def requestNLText(cls):
        state = cls.getState()
        if not (state & cls.S_RECEIVED_SIGNAL):
            raise Exception("Signal file was not provided, cannot return NL text!")

        signalFilePath = cls.getConfig(cls.P_LAST_SIGNAL_FILE)

        path, filename = os.path.split(signalFilePath)
        outputFile = os.path.join(path, cls.basenameExtractorRegex.sub(r"\g<basename>.feature", filename))
        
        # print("SignalToNL: ", signalFilePath)
        signalToNL(signalFilePath)

        cls.removeState(cls.S_RECEIVED_SIGNAL)

        responseMode = cls.getConfig(cls.P_RESPONSE_MODE)
        if responseMode == cls.PV_RESPONSE_FILE:
            response = send_file(outputFile)
        elif responseMode == cls.PV_RESPONSE_JSON:
            with open(outputFile, "r") as nlTextFile:
                nlText = nlTextFile.read()

            response = {
                "language":"en<OR>cz",
                "description": "This feature is not implemented, yet! Use RESPONSE_MODE: FILE instead.",
                "background": "background/context text",
                "scenarios": nlText
            }
        else:
            response = {"Message": "Request mode is set to an incorrect value!"}
        return response

    @classmethod
    def receiveNLText(cls, data):
        state = cls.getState()
        if state & (cls.S_RECEIVED_SIGNAL | cls.S_RECEIVED_AUDIO):
            raise Exception("Different data type already received!")

        cls.addState(cls.S_RECEIVED_NL_TEXT)

        basepath = os.path.join(cls.dataFolder, cls.generateDatestring())
        nlFilePath = f"{basepath}.feature"
        signalFilePath = f"{basepath}_signals.feature"

        # print("nlToSignal: ", basepath, data)
        nlToSignal(basepath, data)
        
        cls.addState(cls.S_FINISHED_PROCESSING_NL_TEXT)
        cls.setConfig(cls.P_LAST_NL_TEXT_FILE, nlFilePath)
        cls.setConfig(cls.P_LAST_SIGNAL_FILE, signalFilePath)

        # TODO: return possible errors
        return {
            "info": "Processing done",
            "lines": [],
            "error_lines": [],
            "error_hints": [],
            "errors": False
        }

    @classmethod
    def receiveAudio(cls, audioPath, language):
        state = cls.getState()
        if state & (cls.S_RECEIVED_SIGNAL | cls.S_RECEIVED_NL_TEXT):
            raise Exception("Different data type already received!")

        if language == "en":
            lang = "en-US"
        elif language == "en":
            lang = "cs-CZ"
        transcriptPath = '.'.join([audioPath[:audioPath.find(".")], ".txt"])

        # print("speechToText: ", audioPath, lang)
        speechToText(audioPath, lang)

        return send_file(transcriptPath)

    @classmethod
    def requestRobotPrograms(cls, language):
        if not language:
            language = cls.language

        with open(cls.robotProgramsPath.format(language), 'r') as robotProgramFile:
            robotPrograms = robotProgramFile.read()

        return robotPrograms

    @classmethod
    def receiveRobotPrograms(cls, programs, language):
        if not language:
            language = cls.language

        with open(cls.robotProgramsPath.format(language), 'w') as robotProgramFile:
            robotProgramFile.writelines(programs)

    # =====>>> RESTful API Helper functions <<<===== #
    @classmethod
    def canRequestSignal(cls):
        """Checks whether it is possible to request a signal file.
        """
        state = cls.getState()
        return state & cls.S_FINISHED_PROCESSING_NL_TEXT

    @classmethod
    def canRequestNLText(cls):
        """Checks whether it is possible to request a NL text file.
        """
        state = cls.getState()
        return state & cls.S_RECEIVED_SIGNAL



"""
Testing of functionalities from class RawNLParser
"""


from gherkan.encoder.RawNLParser import RawNLParser


text_raw = "scenario given that robot r3 picks up box and robot r3 collects cubes. scenario when robot r3 picks up cube one, robot r3 puts " \
           "the cube on the shuttle XY on position two. scenario when robot r4 puts cubes on chassi, then cube five is picked up by robot r2"

raw_parser = RawNLParser()
raw_parser.parse(text_raw)

## Generation of list of robot actions. Make_new = True - overwrites the current RobotPrograms_en, False only updates it
raw_parser.generate_program_dict(make_new=False)




from gherkan.decoder.SignalParser import SignalParser

if __name__ == "__main__":
    # Parse statement into the SignalStatementTree
    sp = SignalParser()
    signalTree = sp.parseStatement("(robotN1ProgramNumber == 1 && edge(robotN3ProgramStart, 1)) && (robotN1ProgramNumber == 1 && edge(robotN3ProgramStart, 1))")
    print(signalTree)




from flask import jsonify, request
from flask_restful import Resource, reqparse
from .. import API_FSA

class Actions(Resource):
    parser = reqparse.RequestParser()

    def get(self, language=''):
        return API_FSA.requestRobotPrograms(language)

    def post(self, language=''):
        API_FSA.receiveRobotPrograms(request.form["data"], language)
        return {"OK": True}




import requests

"""Testing REST API

url - 

"""

address = '127.0.0.1'
port = 5000

simple_endpoints = ['signal_map', 'signal_map/cz', 'signal_map/en', 'actions']

complex_endpoints = ['nltext', 'signals']
complex_endpoint_requests = [
    ''
]

for endpoint in simple_endpoints:
    print("----------------------------------->")
    print(f"Testing simple GET request to endpoint '{endpoint}'")
    response = requests.get(f"http://{address}:{port}/{endpoint}").json()
    print(response)
    print("<-----------------------------------")

for endpoint, request in zip(complex_endpoints, complex_endpoint_requests):
    print("----------------------------------->")
    print(f"Testing request to endpoint '{endpoint}'")
    response = requests.get(f"http://{address}:{port}/{endpoint}").json()
    print(response)
    print("<-----------------------------------")



# coding: utf-8
"""
Class FillActions turns the filled object SignalPhrase into a natural language phrase, which is then returned
as the filled fields niceStrEN and niceStrCZ in the SignalPhrase.
"""
import yaml, re, json
import os
from termcolor import colored
from gherkan.utils.word_morpho import Finetuner
import gherkan.utils.constants as c
from pattern.en import conjugate, PARTICIPLE

CZ_PROGRAM_DICT_DIR = os.path.join(c.PROJECT_ROOT_DIR, "gherkan/utils", "RobotPrograms_cz.json")
EN_PROGRAM_DICT_DIR= os.path.join(c.PROJECT_ROOT_DIR, "gherkan/utils", "RobotPrograms_en.json")
GENERAL_YAML_DIR = os.path.join(c.PROJECT_ROOT_DIR, "gherkan/utils", "en-cz_phrases.yaml")

class FillActions:
    def __init__(self):
        self.finetune = Finetuner()
        self.lang = None

    def ParseMainAction(self, data):
        """ Construct basic NL phrase from given template
            Input format of data:
            data = [Name, Vars, Values, TempToNL]
            Vars = list of strings; specific values to fill into template
            Values = list of integers; values for given Types
            TempToNL = string; specific details for given template
        """
        self.data = self.convert_to_dic(data)
        self.lang = self.data["language"]
        varz = self.data["vars"]
        values = self.data["values"]
        phrase = self.data["tempToNL"][-1]
        for key in varz:  # separate actor names (shuttleXY > shuttle XY)
            if key in ["subject", "object"]:
                if not varz[key].isupper():
                    string = str(varz[key][1:])  # we leave out the first letter in case it is upper
                    actor = (re.findall('([A-Z0-9].*)', string))
                    if actor:
                        varz[key] = varz[key].replace(actor[0], " {}".format(actor[0]))
                    else:
                        pass
        for key, v in values.items():  # try to match the values with .json
                matched = self.MatchAction(varz['subject'], values[key])
                if matched:
                   values[key] = matched
                   to_replace = re.compile(r"\(\?P\<{}\>[^\)]+\)".format(key))
                   phrase = re.sub(to_replace, values[key], phrase)
                else:
                    phrase = self.ParseBool(phrase, values[key])
        for key, value in varz.items():  # match varz with template
            idx = list(varz).index(key)
            value = self.MatchGeneral(value)
            if self.data["inclinations"][idx] != 0:
                value = self.finetune.conjugate(self.lang, value, self.data["inclinations"][idx])
            to_replace = re.compile(r"\(\?P\<{}\>[^\)]+\)".format(key))
            phrase = re.sub(to_replace, value, phrase)
        verb = self.finetune.find_verb(self.lang, phrase)
        if verb:
          verb_conj = self.finetune.conjugate(self.lang, verb[0], 3) # conjugate verb to third person
          phrase = re.sub(verb[0], verb_conj, phrase)
        for word in phrase.split():
          word_new = self.MatchGeneral(word)  # translate words to czech
          phrase = re.sub(r'\b({})\b'.format(word), word_new, phrase)
        phrase = self.finetune.strip_extra_spaces(phrase)
        phrase = self.finetune.align_adjectives(self.lang, phrase)

        if phrase:
            data.__dict__["niceStr"] = phrase
        else:
            data.__dict__["niceStr"] = self.data["niceStr"]
            print(colored("Failed to create NL description for following actions: {}".format(self.data["niceStr"]), "yellow"))


    def ParseBool(self, phrase, value):
        """ Bool 1 is positive, Bool 0 negative (sth does not happen) """
        exceptions = ["be", "can", "will", "must"]
        if value:
            phrase_bool = phrase
        elif not value:
            verb = self.finetune.find_verb(self.lang, phrase)
            if verb:
                if self.lang == "cs":
                    phrase_bool = re.sub(verb[0], "".join(["ne", verb[0]]), phrase)
                elif self.lang == "en":
                    verb_lemma = self.finetune.lemmatize(self.lang, verb[0])
                    if verb_lemma in exceptions:
                        phrase_bool = re.sub(r'\b{}\b'.format(verb[0]),  " ".join([verb_lemma, "not"]), phrase)
                    else:
                        phrase_bool = re.sub(r'\b{}\b'.format(verb[0]), " ".join(["does not", verb_lemma]), phrase)
            else:
                phrase_bool = phrase
        return phrase_bool

    def ParseForce(self, phrase, value):
        if value:
            if self.lang == "en":
                phrase = " ".join(["force", phrase])
            elif self.lang == "cs":
                phrase = " ".join(["vynuť", phrase])
        elif not value:
            phrase = phrase
        return phrase

    def convert_to_dic(self, lists):
        data = lists.__dict__
        return data

    def strip_extra_spaces(self, text):
        stripped_spaces = re.sub(' +', ' ', text)
        stripped_text = stripped_spaces.strip()
        return stripped_text

    def start_stop(self, phrase, state):
        state = self.finetune.conjugate(self.lang, state.lower(), 3)
        verb = self.finetune.find_verb(self.lang, phrase)
        verb_new = verb[0]
        if verb:
            if self.lang == "cs":
                verb_new = self.finetune.lemmatize(self.lang, verb[0])
                print(verb_new)
            elif self.lang == "en":
                verb_new = conjugate(verb[0], tense=PARTICIPLE, parse=True)
            phrase = re.sub(verb[0], " ".join([state, verb_new]), phrase)
            return phrase
        else:
            return phrase

    def MatchAction(self, actor, prog):
        """ Load robot actions and return sring for given program number """
        if self.lang == "en":
            with open(EN_PROGRAM_DICT_DIR, 'r') as stream:
                action_list = json.load(stream)
                try:
                  action = action_list[actor.lower()][str(prog)]
                except:
                    if actor.lower() in action_list:
                       action = "execute program {}".format(prog)
                       print(colored("Warning: Did not find program {} for robot {} in English".format(prog, actor), "yellow"))
                    else:
                        return False

        elif self.lang == "cs":
            with open(CZ_PROGRAM_DICT_DIR, 'r') as stream:
                action_list = json.load(stream)
                try:
                   action = action_list[actor.lower()][str(prog)]
                except:
                    if actor.lower() in action_list:
                        action = "vykonat program {}".format(prog)
                        print(colored("Warning: Did not find program {} for robot {} in Czech".format(prog, actor), "yellow"))
                    else:
                        return False
        return action

    def MatchGeneral(self, word):
        """ Return nice string for basic form of a word """
        with open(GENERAL_YAML_DIR, 'r') as stream:
            phrase_list = yaml.load(stream)
            try:
              phrase = phrase_list[self.lang][word.lower()]
            except:
                phrase = word
        return phrase



if __name__=="__main__":
    fill_action = FillActions()



from gherkan.containers.StatementTree import StatementTree
from gherkan.decoder.Parser import Parser

from lark import Lark, Token
from gherkan.utils import constants as c

class NLParser(Parser):
    def __init__(self):
        super().__init__()

        self.statement_grammar = """
            ?start: expression

            ?expression: statement
                | "(" expression ")"
                | expression "{AND}" expression      -> and
                | expression "{OR}" expression       -> or
            
            ?statement: /((?!{AND}|{OR}).)+/
            
            %import common.WS_INLINE
            %ignore WS_INLINE
        """

    def determineGrammarByLanguage(self):
        if self.language == c.LANG_EN:
            self.statement_grammar = self.statement_grammar.format(AND="AND", OR="OR")
        elif self.language == c.LANG_CZ:
            self.statement_grammar = self.statement_grammar.format(AND="A", OR="NEBO")
        else:
            raise NotImplementedError("Language {} not recognized".format(self.language))

    def parseStatement(self, statement: str):
        self.determineGrammarByLanguage()

        parser = Lark(self.statement_grammar, parser='earley', ambiguity='resolve', propagate_positions=True)
        tree = parser.parse(statement)

        if type(tree) == Token:
            print(tree)
        else:
            print(tree.pretty())

        st = StatementTree(statement)
        st.buildFromNLTree(tree)
        
        print(st)

        return st



# coding: utf-8
# from googletrans import Translator
import majka, re, os, imp
from pattern.en import conjugate, parsetree
from termcolor import colored
import spacy, unicodedata

class Finetuner():
    def __init__(self):
        self.morph_lwt = majka.Majka(os.path.join(imp.find_module("gherkan")[1], "utils/majka.l-wt"))
        self.morph_wlt = majka.Majka(os.path.join(imp.find_module("gherkan")[1], "utils/majka.w-lt"))

    def conjugate(self, lang, word, person, number="singular"):
        """  Returns provided noun in given grammatical case / Vrátí podstatné jméno vyskloňované v daném pádu """
        declenation = ""
        if lang == "en":
           try:
            if number == "singular":
                tag = str(person) + "sg"
                declenation = conjugate(word, tag)
            elif number == "plural":
                tag = str(person) + "pl"
                declenation = conjugate(word, tag)
           except:
               declenation = word
        elif lang == "cs":
           try:
            self.morph_lwt.flags |= majka.ADD_DIACRITICS
            analysis = self.morph_lwt.find(word)
            if analysis == []:
                declenation = word
            if number == "singular":
                for n in range(len(analysis)):
                    if 'singular' in analysis[n]['tags']:
                        declen = [analysis[n]['lemma'], analysis[n]['tags']]
                        if "person" in declen[1]:
                            if declen[1]["person"] == int(person) and declen[1]["singular"] is True and declen[1]["negation"] is False:
                                declenation = declen[0]
            else:
                for n in range(len(analysis)):
                    if 'singular' in analysis[n]['tags']:
                        declen = [analysis[n]['lemma'], analysis[n]['tags']]
                        if "person" in declen[1]:
                            if declen[1]["person"] == int(person) and declen[1]["singular"] is False and declen[1]["negation"] is False:
                                declenation = declen[0]
           except:
               declenation = word
        if not declenation:
            declenation = word
        if word == "nebýt" and person == 3:   # majka does not work for this case
            if number == "singular":
                declenation = "není"
            else:
                declenation = "nejsou"
        return declenation

    def start_sentence_upper(self, text):
        """Changes the letters at the beginning of sentences into upper"""
        return re.sub("(^|[.?!])\s*([a-zA-Z])", lambda p: p.group(0).upper(), text.lower())

    def strip_extra_spaces(self, text):
        stripped_spaces = re.sub(' +', ' ', text)
        stripped_text = stripped_spaces.strip()
        return stripped_text

    def lemmatize(self, lang, word):
        """  Returns lemma of given word form / Vrátí základní podobu slova (např. infinitiv)"""
        if lang == "cs":
            self.morph_wlt.flags |= majka.ADD_DIACRITICS  # find word forms with diacritics
            self.morph_wlt.tags = False  # return just the lemma, do not process the tags
            self.morph_wlt.first_only = True  # return only the first entry
            output = self.morph_wlt.find(word)
            lemma = output[0]['lemma']
        else:
            # this version keeps capitals
            nlp = spacy.load('en_core_web_sm', disable=['tagger', 'parser', 'ner'])
            doc = nlp(word)
            for token in doc:
                    lemma = str(token.lemma_)

        return lemma


    def change_case(self, word, case, plural = False):
        """  Returns provided noun in given grammatical case / Vrátí podstatné jméno vyskloňované v daném pádu """
        self.morph_lwt.flags |= majka.ADD_DIACRITICS
        word_form = []
        analysis = self.morph_lwt.find(word)

        if plural == False:
            for n in range(len(analysis)):
                if 'singular' in analysis[n]['tags']:
                    declen = [analysis[n]['lemma'],analysis[n]['tags']['case']]
                    if case in declen:
                       word_form.append(declen[0])
                else:
                    pass
        else:
            for n in range(len(analysis)):
                if 'plural' in analysis[n]['tags']:
                    declen = [analysis[n]['lemma'], analysis[n]['tags']['case']]
                    if case in declen:
                        word_form.append(declen[0])
            else:
                pass

        return word_form[0]

    def make_negative(self, lang, word):
        if lang == "cs":
            self.morph_wlt.flags |= majka.ADD_DIACRITICS
            self.morph_wlt.tags = False
            self.morph_wlt.first_only = False
            analysis = self.morph_wlt.find(word)
        elif lang == "en":
            neg = conjugate(word, negated=True)

    def get_cases(self, word):
        """ Returns all possible cases of the word, in case of verb returns all info / Vrátí možné pády daného slova, u sloves vrátí všechno info"""
        self.morph_wlt.flags |= majka.ADD_DIACRITICS  # find word forms with diacritics
        output = self.morph_wlt.find(word)

        cases = []
        if "substantive" in output[0]['tags']['pos'] or "adjective" in output[0]['tags']['pos']:
            for n in range(len(output)):
                 case = output[n]['tags']['case']
                 if "singular" in output[n]['tags']:
                  number = "singular"
                 else:
                  number = "plural"
                 cases.append([case, number])

        elif "verb" in output[0]['tags']['pos']:
            cases = output

        return cases


    def align_adjectives(self, lang, sentence):
        word_list = sentence.split()
        self.morph_wlt.flags |= majka.ADD_DIACRITICS
        if lang =="cs":
            for word in word_list:
                analysis = self.morph_wlt.find(word)
                if analysis == []:
                   continue
                else:
                    if analysis[0]["tags"]["pos"] == "substantive":
                        config = analysis[0]["tags"]
                        config["negation"] = False
                        del config["pos"]
                        if 'plural' in config:
                            del config["plural"]
            if config:
                verb = self.find_verb(lang, sentence)
                try:
                   analysis = self.morph_wlt.find(verb[0])
                   config["singular"] = analysis[0]["tags"]["singular"]
                except:
                    pass
                for word in word_list:
                    analysis = self.morph_lwt.find(word)
                    if analysis:
                        if analysis[0]["tags"]["pos"] == "adjective":
                            for dict in analysis:
                                if all(item in dict["tags"].items() for item in config.items()):
                                    aligned_adj = dict["lemma"]
                                    sentence = re.sub(word, aligned_adj, sentence)
        else:
            pass
        return sentence

    def remove_diacritics(self, text):
        """
        Returns a string with all diacritics (aka non-spacing marks) removed.
        For example "Héllô" will become "Hello".
        Useful for comparing strings in an accent-insensitive fashion.
        """
        normalized = unicodedata.normalize("NFKD", text)
        return "".join(c for c in normalized if unicodedata.category(c) != "Mn")

    def find_verb(self, lang, sentence, head_only=True):
        verb = []
        word_list = sentence.split()
        if lang =="cs":
            for word in word_list:
                analysis = self.morph_wlt.find(word)
                if analysis == []:
                   continue
                else:
                    if analysis[0]["tags"]["pos"] == "verb":
                        verb.append(word)
                    else:
                        continue
        elif lang == "en":
            t = parsetree(sentence, tokenize=True,  tags=True)
            excepts = ["start", "stop", "end", "finish"]
            for sentence in t:
              try:
                  chunk = sentence.verbs[0]
                  check = False
                  for x in excepts:
                      if x in chunk.string:
                        word = re.findall(r'\w*{}\w*'.format(x), chunk.string)
                        if word:
                          verb.append(word[0])
                          check = True
                  if not check:
                     if head_only:
                       verb.append(chunk.head.string)
                     else:
                       verb.append(chunk.string)
              except:
                  print(colored("Warning: No verb detected in {}".format(sentence), "yellow"))

        return verb

    # def translate(self, phrase, lang = "cz-en"):
    #     """ Translates a word or phrase from czech to english (lang = "cz-en") or from english to czech (lang = "en-cz") """
    #     translator = Translator()
    #     if lang == "cz-en":
    #       result = translator.translate(phrase, src="cs", dest='en')
    #     elif lang == "en-cz":
    #       result = translator.translate(phrase, src="en", dest='cs')
    #
    #     translation = result.text
    #     return translation



if __name__ == "__main__":
 adjust = Finetuner()



import unittest
"""
This code will gather and run all the tests in this directory. Files with tests must start with the "test_" prefix.
"""

loader = unittest.TestLoader()
testSuite = loader.discover(".")

testRunner = unittest.TextTestRunner(verbosity=2)
testRunner.run(testSuite)



FEATURE = "Feature"
BACKGROUND = "Background"
SCENARIO = "Scenario"
GIVEN = "Given"
WHEN = "When"
THEN = "Then"
AND = "And"
BUT = "But"

FEATURE_CS = "Požadavek"
BACKGROUND_CS = "Kontext"
SCENARIO_CS = "Scénář"
GIVEN_CS = "Pokud"
WHEN_CS = "Když"
THEN_CS = "Pak"
AND_CS = "A"
BUT_CS = "Ale"


import os
import gherkan.utils.constants as c

from gherkan.flask_api.raw_text_to_signal import process

request = {
    "feature" : "robot R1",
    "feature_desc" : "lorem ipsum",
    "background" : "given Line is On",
    "text_raw" : "scenario As soon as shuttle XY is in the station ZX, then robot r1 picks "
                 "up cube one. scenario when robot r3 picks up cube one, robot r2 assembles the product.",
    "language" : "en"
}
base_path = os.path.join(c.DATA_DIR, "output", "raw_out")

process(base_path, request)

nl_file_path = base_path + ".feature"
signal_file_path = base_path + "_signals.feature"

print("\n\n--- NL FILE ---")
with open(nl_file_path, "rt") as f:
    text = f.read()
    print(text)

print("\n\n--- SIGNAL FILE ---")
with open(signal_file_path, "rt") as f:
    text = f.read()
    print(text)


from gherkan.containers.signal_batch import SignalBatch
from gherkan.containers.StatementTreeNode import StatementTreeBinaryOperatorNode, \
    StatementTreeNode, StatementTreeOperandNode, StatementTreeMergedNode
from gherkan.containers.StatementTree import StatementTree
from gherkan.processing.TreeProcessor import TreeProcessor
import gherkan.utils.gherkin_keywords as g

class NLFileWriter():
    """ TODO docs """

    def __init__(self, signalBatch : SignalBatch):
        self.signalBatch = signalBatch
        self.outLines = []
        self.language = signalBatch.language

        self.encode(signalBatch)

    def encode(self, signalBatch: SignalBatch):
        self.outLines.append("# language: {}\n\n".format(self.language))
        self.outLines.append("{}: {}\n".format(g.FEATURE, signalBatch.name))
        self.outLines.append("  {}\n".format(signalBatch.desc))

        if signalBatch.context:
            self.outLines.append("{}:\n".format(g.BACKGROUND))
            self.outLines.append("  {} {}\n\n".format(g.GIVEN, self.tree_to_str(signalBatch.context)))

        for scenario in signalBatch.scenarios:
            self.outLines.append("{}: {}\n".format(g.SCENARIO, scenario.name))

            for tree in scenario.givenStatements:
                self.outLines.append("  {} {}\n".format(g.GIVEN, self.tree_to_str(tree)))
            for tree in scenario.whenStatements:
                self.outLines.append("  {} {}\n".format(g.WHEN, self.tree_to_str(tree)))
            for tree in scenario.thenStatements:
                self.outLines.append("  {} {}\n".format(g.THEN, self.tree_to_str(tree)))

    def write(self, outputFilePath: str):
       with open(outputFilePath, "w") as out:
           out.writelines(self.outLines)

    def tree_to_str(self, tree: StatementTree):
        tp = TreeProcessor(self.language)

        # loads yaml file with templates
        tp.load_templ_dic('utils/templates_dic.yaml')
        tree.root = tp.process_tree(tree.root, TreeProcessor.Direction.SIGNAL_TO_NL)

        return self.node_to_str(tree.root)

    def node_to_str(self, node: StatementTreeNode):
        if type(node) == StatementTreeBinaryOperatorNode:
            return self.operator_to_str(node)
        elif type(node) == StatementTreeOperandNode or \
              type(node) == StatementTreeMergedNode:
            return self.operand_to_str(node)

    def operator_to_str(self, node: StatementTreeBinaryOperatorNode):
        return ("({} {} {})".format(
            self.node_to_str(node.lchild),
            node.kind.upper(),
            self.node_to_str(node.rchild))
        )

    def operand_to_str(self, node: StatementTreeOperandNode):
        return "{}".format(node.data.variableNL)


import os
import json
import string
import en_coref_md
from pattern.en import parsetree, conjugate, lemma
from pattern.search import Pattern
import itertools
import spacy
import re
import gherkan.utils.constants as c
import gherkan.utils.gherkin_keywords as g
from gherkan.encoder.NL2Temp import NLTempMatcher


#TODO: pass language from Raw Text
LANG = "en"

class RawNLParser:
    def __init__(self):
        self.nlp = spacy.load('en_core_web_sm')
        self.text_raw = ""
        self.robot_actions = []
        self.text_lines_normalized = []
        self.sentences = []
        self.cleaner = NLTempMatcher(LANG)
        self.subj_of_interest = "robot"

    def parse(self, text_raw: str):
        self.text_raw = text_raw
        scenarios = self.separate_scenarios(text_raw)

        for scenario in scenarios:
            print("Scenario: " + scenario)
            self.text_lines_normalized.append("Scenario: {}".format(scenario))
            text = self.replace_synonyms(scenario)
            self.collect_subject_actions(text)
            separated_text = self.separate_sentences(text)
            for phrase in separated_text:
                clean = self.strip_extra_spaces(phrase)
                print(clean)
                self.text_lines_normalized.append(clean)

        print("Actions connected with {}s: ".format(self.subj_of_interest), self.robot_actions)

    def process_sentence(self, sentence):
        """ Aimed for processing of Background. Will not sort into template"""
        text = self.replace_synonyms(sentence)
        self.collect_subject_actions(text)
        given = list(filter(None, re.split(r'^\s*given', text, re.IGNORECASE)))
        lemmatized = self.strip_extra_spaces("given" + self.lemmatize(given[0]))
        return lemmatized

    def get_text_lines(self):
        print(self.text_lines_normalized)
        return self.text_lines_normalized

    def replace_synonyms(self, command):
        """place words in command with synonyms defined in synonym_file"""
        synonyms_filename = os.path.join(
            c.PROJECT_ROOT_DIR, "gherkan", "utils", "synonyms.json")
        with open(synonyms_filename) as f:
            synonyms = json.load(f)
        for key in synonyms["en"]:
            for phrase in synonyms["en"][key]:
                if phrase in command.lower():
                    src_str = re.compile(phrase, re.IGNORECASE)
                    command = src_str.sub(key, command)
        command_ready = command.replace(" todelete", "")
        # print("Replaced: " + self.command_ready)
        return command_ready

    def separate_sentences(self, text):
        """ Separate the sentences according to given keywords Given, When, Then """
        separated = []
        self.sentences = []
        templates = ["Given ", "When ", "Then "]
        for word in templates:
            separated = self.find_patterns(text, word)

        return separated

    def find_patterns(self, text, temp_word):
        """ Detects sentences starting with temp_word and end with "," or "."
            Sentences without any keyword are detected as Then"""
        detected_temp = re.findall(
            r'(?<=(?i)' + re.escape(temp_word) + r')+([^,]+)', text)
        rest = []
        for x in detected_temp:
            rest = re.split(x, text)[1]
            text = rest
            rest = rest.translate(str.maketrans('', '', string.punctuation))
            rest = self.strip_extra_spaces(rest)
        for x in detected_temp:
            new_match = temp_word + x
            new_match = new_match.replace(" and ", " AND ")
            new_match = new_match.translate(
                str.maketrans('', '', string.punctuation))
            new_match = self.strip_extra_spaces(new_match)
            self.sentences.append(new_match)
        if rest:
            rest = "Then " + re.compile("then ", re.IGNORECASE).sub("", rest)
            self.sentences.append(rest)

        for a, b in itertools.combinations(self.sentences, 2):
            if a in b:
                self.sentences.remove(a)
            elif b in a:
                self.sentences.remove(b)

        return self.sentences

    def generate_program_dict(self, make_new=True):
        """ make_new - creates a new dict and replaces old one. If False, programs are appended to current dict """
        program_dict = {}
        dict_path = os.path.join(
            c.PROJECT_ROOT_DIR, "gherkan", "utils", "RobotPrograms_en.json")
        if self.robot_actions:
            if not make_new:
                with open(dict_path, 'r') as f:
                    program_dict = json.load(f)
                    f.close()
            for action in self.robot_actions:
                tree = parsetree(action, tokenize=True,
                                 tags=True, chunks=True, relations=True)
                for sentence in tree:
                    actor = str(sentence.subjects[0]).lower()
                action_lemma = self.lemmatize("".join(action.rsplit(actor)))
                if actor in program_dict.keys():
                    is_unique = True
                    # find if the action is already in list
                    for key in program_dict[actor]:
                        if self.strip_extra_spaces(action_lemma) in program_dict[actor][key]:
                            is_unique = False
                    if is_unique:
                        new_key = len(program_dict[actor]) + 1
                        program_dict[actor].update(
                            {new_key: self.strip_extra_spaces(action_lemma)})
                else:
                    program_dict[actor] = {
                        1: self.strip_extra_spaces(action_lemma)}
            with open(dict_path, 'w') as f:
                json.dump(program_dict, f)
            print("Program list: {}".format(program_dict))
            print("Programs saved to {}".format(dict_path))
        else:
            raise Exception(
                "No robot actions found. Have you parsed the text first?")

    def strip_extra_spaces(self, text):
        stripped_spaces = re.sub(' +', ' ', text)
        stripped_text = stripped_spaces.strip()
        return stripped_text

    def collect_subject_actions(self, text):
        """ Detect any actions performed by a robot """
        tree = parsetree(text, tokenize=True, tags=True,
                         chunks=True, relations=True)
        sentences = self.separate_compound(tree)
        for x in sentences:  # now we select sentences with robot subjects or passive form
            sent_parsed = parsetree(x, chunks=True, relations=True)
            for sentence in sent_parsed:
                for chunk in sentence.chunks:
                    if chunk in sentence.subjects:
                        if self.subj_of_interest in chunk.string:  # active verb form
                            sentence = sentence.slice(
                                chunk.start, sentence.stop)
                            sent = (sentence.string).translate(
                                str.maketrans('', '', string.punctuation))
                            bool = self.cleaner.get_negative(sent)
                            force = self.cleaner.get_force(sent)
                            sentence_new = self.cleaner.get_cleanphr(sent, bool, force)
                            self.robot_actions.append(sentence_new)
                if " ".join(["by", self.subj_of_interest]) in sentence.string.lower():  # handle passive form
                    verbs, nouns = ([] for i in range(2))
                    verb_object = (sentence.string).split(" by ")[0]
                    tree = parsetree(verb_object, lemmata=True)
                    subject = (sentence.string).split(" by ")[1]
                    for part in tree:
                        for chunk in part.chunks:
                            if chunk.type == "VP":
                                for word in chunk.words:
                                    verbs.append(
                                        (conjugate(word.string, tense="INFINITIVE")))
                                if "be" in verbs:
                                    verbs.remove("be")
                                rest = verb_object.replace(chunk.string, "")
                    sent = subject + " " + ' '.join(verbs) + " " + rest
                    sent = sent.translate(
                        str.maketrans('', '', string.punctuation))
                    bool = self.cleaner.get_negative(sent)
                    force = self.cleaner.get_force(sent)
                    sentence_new = self.cleaner.get_cleanphr(sent, bool, force)
                    self.robot_actions.append(sentence_new)
        # now remove duplicit actions:
        for a, b in itertools.combinations(self.robot_actions, 2):
            if a in b:
                self.robot_actions.remove(a)
            elif b in a:
                self.robot_actions.remove(b)

    def separate_compound(self, tree):
        """ Separate individual sentences in a compound sentence"""
        sentence_num = 1  # we always have at least one sentence
        sentence_start = 0
        sentences = []
        for sentence in tree:
            for chunk in sentence.chunks:  # first we split compound sentences according to chunk relations
                if chunk.relation is not None and chunk.relation is not sentence_num:
                    new_sentence = sentence.slice(sentence_start, chunk.start)
                    sentences.append(new_sentence.string)
                    sentence_start = chunk.start
                    sentence_num = chunk.relation
            sentence_new = sentence.slice(sentence_start, sentence.stop)
            sentences.append(sentence_new.string)
        return sentences

    def lemmatize(self, text):
        lemmas = ""
        # this version keeps capitals
        tree = parsetree(text, tokenize=True)
        skip_list = [",", ".", ";"]
        temp_words = [g.GIVEN, g.WHEN, g.THEN]

        for sentence in tree:
            for word in sentence:
              if word.string in temp_words:
                 lemmas += word.string
              else:
                if word.string in skip_list:
                    pass
                else:
                    lemmas += " "
                lemmas += lemma(word.string)

        lemmatised = lemmas
        # print(lemmatised)

        return lemmatised

    def solve_corefs(self, text):
        print("Loading Neuralcoref module...")
        nlc = en_coref_lg.load()
        dok = nlc(text)
        if dok._.has_coref == 1:
            replaced_corefs = dok._.coref_resolved
        else:
            replaced_corefs = text
        print("Done")
        # print("Corref: " + replaced_corefs)

        return replaced_corefs

    def separate_scenarios(self, text):
        text = text
        split = text.split("scenario ")
        split = list(filter(None, split))
        # print("Scenarios: ", split)

        return split

    def replace(self, string, substitutions):
        substrings = sorted(substitutions, key=len, reverse=True)
        regex = re.compile('|'.join(map(re.escape, substrings)))

        return regex.sub(lambda match: substitutions[match.group(0)], string)



# -*- coding: utf-8 -*-


class DictionaryMapper():
    """
    This is essentially shaped down Translator that works with Python dictionaries (instead of dicts from a YAML file).
    """

    def __init__(self, dictionary):
        self.dictionary = dictionary
        self.lookup = {}    # inverted dictionary
        self.conseq = {}    # list of words after which other words can appear
        self.alternativeWords = []

        # word = target word into which the alternatives will be translated
        # alts = alternatives/atlernative words
        for (word, alts) in dictionary.items():
            for alt in alts:
                if " " in alt:
                    altsplit = alt.split(" ")
                    self.__addConsecutiveAlts(altsplit, self.conseq)
                if alt in self.lookup:
                    raise Exception("Error: ambiguous alternative. The word {} was provided as an alternative to {} but it was already assigned as an alternative to {}!".format(alt, word, self.lookup[alt]))
                self.lookup[alt] = word

        self.alternativeWords = frozenset(self.lookup.keys())

    def __addConsecutiveAlts(self, altRemainder, conseq):
        """ Recursion warning - this function is awful and attempt to understand it requires high amounts of caffeine and carbohydrates
        But essentially, it just traverses a nested dictionary and tries to put words in the right place to create
        chains of words that will be used when parsing phrases in the sentences.
        """
        alt = altRemainder[0]
        nextAlt = altRemainder[1]
        lastPair = len(altRemainder) == 2
        if alt in conseq:
            wcsq = conseq[alt]
            if isinstance(wcsq, str):
                if wcsq == nextAlt:
                    if lastPair:
                        pass  # this should not happen as it means the same phrase was added twice
                    else:
                        conseq[alt] = {wcsq: {".": None}}
                else:
                    if lastPair:
                        conseq[alt] = {wcsq: ".", nextAlt: "."}
                    else:
                        conseq[alt] = {wcsq: ".", nextAlt: {}}
            else:  # wcsq is a dictionary
                if lastPair:
                    if nextAlt in wcsq:
                        if isinstance(wcsq[nextAlt], str):
                            wcsq[nextAlt] = {wcsq[nextAlt]: ".", ".": None}
                        else:
                            wcsq[nextAlt]["."] = None
                    else:
                        wcsq[nextAlt] = "."
        else:  # alt is not in conseq
            if lastPair:
                conseq[alt] = nextAlt
            else:
                conseq[alt] = {}

        if not lastPair:
            self.__addConsecutiveAlts(altRemainder[1:], conseq[alt])

    def __checkConsecutiveWords(self, sentRemainder, conseq):
        """ Recursion warning - this function is awful and attempt to understand it requires high amounts of caffeine and carbohydrates
        But all it does is it tries to traverse the nested dictionary created by the "__addConsecutiveAlts" function
        to properly parse the sentence and attempt to find predefined phrases.
        """
        w = sentRemainder[0]
        if conseq is None or not sentRemainder:
            return []  # if there are no words remaining in the sentence - this should only happen in case of error
        elif isinstance(conseq, str):
            if w == conseq:
                return [conseq]  # if the consecutives is just a single word, then return that word and end
            elif '.' == conseq:
                return True  # current word is not in consequtives but the previous word can be terminal
            else:

                return []
        else:  # else, there are some consecutives to be processed
            followers = []  # words following this word in the phrase
            if w in conseq:
                if len(sentRemainder) > 1:
                    followers = self.__checkConsecutiveWords(sentRemainder[1:], conseq[w])
                elif '.' in conseq[w]:  # dot in the consecutives dictionary means this word can be terminal
                    return [w]  # valid last word in the phrase (not necessarily in the sentence)
                else:
                    return []
            elif '.' in conseq:
                return True  # current word is not in consequtives but the previous word can be terminal
            else:
                return []

            if not followers:
                return []
            elif isinstance(followers, list):
                return [w] + followers
            else:
                return [w]  # the word has no followers in the sentence but was recognized as a valid terminal word

    def filter(self, words):
        """
        Translates the input and filters empty results.

        Parameters
        ----------
        words : str OR list
            A single word or a list of words.

        Returns
        -------
        list:
            Returns a list of translated words and removes empty entries (i.e., entries without corresponding translations in the dictionary).
        """
        # Add ordinal numbers to the list and filter out the empty values.
        return list(filter(lambda ix: ix[1], enumerate(self.map(words))))

    def map(self, words):
        """
        Translates the input.

        Parameters
        ----------
        words : str OR list
            A single word or a list of words.

        Returns
        -------
        list:
            Returns a list of translated words. Returns an empty string for words missing from the dictionary.
        """
        if isinstance(words, str) and " " in words:
            return self.map(words.split(" "))
        elif isinstance(words, list):
            sentence = []
            n = len(words)
            i = 0
            while i < n:
                w = words[i]
                if w in self.conseq and i < n - 1:  # checks for alternative multiwords phrases (instead of a single word)
                    phrase = self.__checkConsecutiveWords(words[i+1:], self.conseq[w])
                    if len(phrase) > 0:  # length check to prevent boolean True from sneaking in, which should only happen in case of an error
                        phrase = [w] + phrase
                        i += len(phrase) - 1
                        w = ' '.join(phrase)
                        sentence.append(self.lookup[w])
                else:
                    sentence.append(self.map(w))
                i += 1
            return sentence
        else:
            if words in self.alternativeWords:
                return self.lookup[words]
            else:
                return ""



import requests
import unittest
import re


class TestSignals(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"
        self.signalFile = """
# language: cs

Požadavek: Robot R1
Robot R1 má na starosti kompletaci skládání Chassi. Robot postupně zbírá kostičky potřebené ke kompletaci.
Kontext:
Pokud LineOn == 1 || (robotR1ProgramNumber == 1 && robotR2ProgramNumber == 1)

Scénář: Normální běh linky - Robot R1 - Skládání Chassi - braní kostičky číslo 1
Když ShuttleXyAtStationYZ == 1 && edge(StationLocked, 1)
Pak   robotR1ProgramNumber == 1 && edge(robotR1ProgramStart, 1)

Scénář: Normální běh linky - Robot R1 - Skládání Chassi - pokládání kostičky číslo 1
Když  robotR1ProgramNumber == 1 && edge(robotR1ProgramEnd, 1)
Pak   robotR1ProgramNumber == 2
"""

    # @unittest.skip("TODO: skipped for debugging")
    def test_post_signal(self):
        data = {'language': 'cs', 'scenarios': self.signalFile}
        response = requests.post(self.url.format("signals"), data=data)
        response = response.json()
        self.assertDictContainsSubset({"OK": True}, response)

    # @unittest.skip("TODO: skipped for debugging")
    def test_post_and_get_nl(self):
        requests.get(self.url.format("reset"))
        correctResponse = """
# language: cs

Feature: Robot R1
  Robot R1 má na starosti kompletaci skládání Chassi. Robot postupně zbírá kostičky potřebené ke kompletaci.
Background:
  Given ((robot R2 dá kostku 1 na podnos AND robot R1 zvedne všechny červené kostky) OR linka jest zapnutá)

Scenario: Normální běh linky - Robot R1 - Skládání Chassi - braní kostičky číslo 1
  When (stanice jest zamčená AND vozíček Xy jest ve stanici YZ)
  Then (robot R1 začne AND robot R1 zvedne všechny červené kostky)
Scenario: Normální běh linky - Robot R1 - Skládání Chassi - pokládání kostičky číslo 1
  When (robot R1 skončí AND robot R1 zvedne všechny červené kostky)
  Then robot R1 zvedne všechny zelené kostky
"""
        wsr = re.compile(u"(^\s+)|(\s+$)", flags=re.MULTILINE | re.UNICODE)
        data = {'language': 'cs', 'scenarios': self.signalFile}
        postResponse = requests.post(self.url.format("signals"), data=data)
        postResponse = postResponse.json()
        self.assertDictContainsSubset({"OK": True}, postResponse)

        response = requests.get(self.url.format("nltext"))
        self.assertMultiLineEqual(wsr.sub(u'', correctResponse), wsr.sub(u"", response.content.decode("utf-8")))




from gherkan.flask_api import gherkan_rest
import sys


if __name__ == "__main__":
    print("Setting up Gherkan API via __main__")
    args = sys.argv[1:] if len(sys.argv) > 1 else None
    gherkan_rest.main(args)


from flask import jsonify, request, Response, make_response
from flask_restful import Resource, reqparse, abort
import re
from .. import API_FSA

class Signals(Resource):

    parser = reqparse.RequestParser()
    parser.add_argument("language", type=str, choices=("en", "cs"), required=True)
    parser.add_argument("background", type=str)
    parser.add_argument("description", type=str)
    parser.add_argument("scenarios", action="append", required=True)

    langDirRegex = re.compile("^.*#\s*lang\w*\s*:.+$", re.IGNORECASE | re.MULTILINE)

    def get(self):
        # Check if NL text was received and processed
        if API_FSA.canRequestSignal():
            # request signal file
            response = API_FSA.requestSignal()
            return response
        else:
            abort(406, message="It is not possible to request a signal file. Most likely no NL text was provided")

    def post(self):
        """
        request = {
                "language": "en<OR>cs",
                "background": "background text",
                "description": "optional description",
                "scenarios": [
                    "scenario1 signal text",
                    "scenario2 signal text"
                ]
        }
        """
        # A signal file arrived
        args = self.parser.parse_args()
        fileText = '\n'.join(args["scenarios"])
        
        # TODO: better text file composition

        if args["background"] is not None:
            fileText = '\n'.join([args["background"], fileText])

        # if self.langDirRegex.search(fileText) is None:
        #     # Add language directive if it is not in the text
        #     fileText = '\n'.join([args["language"], fileText])
        try:
            API_FSA.receiveSignal(fileText)
        except Exception as error:
            abort(406, message="An error occurred while processing the signals: " + str(error))

        return {"OK": True}


from gherkan.utils.dictionaryMapper import DictionaryMapper
from gherkan.utils.dicts import sectionDict
import re
import unittest


class TestDictionaryMapper(unittest.TestCase):

    def setUp(self):
        flatText = u'# language: cs\n\n\n\npozadavek: robot n1\n\n  dil 2 je uchopen robotem n1 a na dil je naneseno lepidlo. v okmziiku, kdy je pripraven dil 1a nebo 1b je dil 2 prilepen k dilu 1. robot odpoji gripper a pripoji svarecku. nasledne k sobe svari dily 2 a 1a/b. po dokonceni svarovani robot odpoji svarecku.\n\n\n\nkontext:\n\n  pokud linkaon == 1\n\n\n\nscenar: zvednuti dilu\n\n  pokud part2ontable == 1 && (robotn1programnumber == 0 || robotn1programnumber == 1 || robotn1programended)\n\n  kdyz  tableathome\n\n  pak   robotn1programnumber == 1 && edge(robotn1programstart, 1)\n\n\n\nscenar: naneseni lepidla\n\n  kdyz  robotn1programnumber == 1 && edge(robotn1programended, 1)\n\n  pak   robotn1programnumber == 2 && edge(robotn1programstart, 1)\n\n\n\nscenar: slepeni dilu\n\n  pokud part1aontable == 1 || part1bontable == 1\n\n  kdyz  tableatfwd == 1\n\n  a     robotn1programnumber == 2 && edge(robotn1programended, 1)\n\n  pak   robotn1programnumber == 3 && edge(robotn1programstart, 1)\n\n\n\nscenar: vymena gripperu za svarecku\n\n  kdyz  robotn1programnumber == 3 && edge(robotn1programended, 1)\n\n  pak   robotn1programnumber == 4 && edge(robotn1programstart, 1)\n\n\n\nscenar: svareni produktu\n\n  kdyz  robotn1programnumber == 4 && edge(robotn1programended, 1)\n\n  pak   robotn1programnumber == 5 && edge(robotn1programstart, 1)\n\n\n\nscenar: vymena svarecky za gripper\n\n  kdyz  robotn1programnumber == 7 && edge(robotn1programended, 1)\n\n  pak   robotn1programnumber == 6 && edge(robotn1programstart, 1)'
        self.candidates = re.findall(r"\n\s*(\w*)", flatText, flags=re.MULTILINE)
        self.correctSections = ['description', '', 'context', 'while', 'scenario', 'while', 'if', 'then', 'scenario', 'if', 'then', 'scenario', 'while', 'if', 'and', 'then', 'scenario', 'if', 'then', 'scenario', 'if', 'then', 'scenario', 'if', 'then']
        self.correctFilteredSections = list(filter(lambda ix: ix[1], enumerate(self.correctSections)))
        self.dm = DictionaryMapper(sectionDict)

    @unittest.skip("Look-up dictionary testing not implemented for the Dictionary Mapper!")
    def test_lookup(self):
        self.assertTrue(False)
        # TODO
        # self.assertDictEqual(self.dm.lookup,
        #                      {}
        #                      )

    def test_map(self):
        self.assertListEqual(self.dm.map(self.candidates), self.correctSections)

    def test_filter(self):
        self.assertListEqual(self.dm.filter(self.candidates), self.correctFilteredSections)




import os
import gherkan.utils.constants as c
from gherkan.flask_api.signal_to_nl import process

# Path to the text file with the signal batch

dir = os.path.join(c.DATA_DIR, 'output')
input_file = os.path.join(dir, 'case_R1NormalSequence2_signals.feature')

process(input_file)

output_file = os.path.join(dir, 'case_R1NormalSequence2.feature')

with open(output_file, "rt") as f:
    text = f.read()
    print(text)


from flask import jsonify, request, Response
from flask_restful import Resource, reqparse, abort
import re
from .. import API_FSA

class Handler(Resource):
    A_RESET = "reset"
    A_SET_RESPONSE_MODE = "set_response_mode"

    def get(self, action, param=None):
        if action == self.A_RESET:
            API_FSA.regenerateConfig()
            API_FSA.setState(API_FSA.S_IDLE)
        elif action == self.A_SET_RESPONSE_MODE:
            API_FSA.setConfig(API_FSA.P_RESPONSE_MODE, param)
        else:
            abort(405, message="Requested action is not defined!")

        return {"OK": True}


from gherkan.containers.signal_batch import SignalBatch
from gherkan.containers.StatementTreeNode import StatementTreeBinaryOperatorNode, \
    StatementTreeNode, StatementTreeOperandNode, StatementTreeMergedNode
from gherkan.containers.StatementTree import StatementTree
from gherkan.processing.TreeProcessor import TreeProcessor

import gherkan.utils.constants as c
import gherkan.utils.gherkin_keywords as g

import warnings

class SignalFileWriter():
    """ TODO docs """

    def __init__(self, nlBatch : SignalBatch):
        self.nlBatch = nlBatch
        self.outLines = []
        self.language = nlBatch.language

        self.encode(nlBatch)

    def encode(self, nlBatch: SignalBatch):
        self.outLines.append("# language: {}\n\n".format(self.language))
        self.outLines.append("{}: {}\n".format(g.FEATURE, nlBatch.name))
        self.outLines.append("  {}\n".format(nlBatch.desc))

        if nlBatch.context:
            self.outLines.append("{}:\n".format(g.BACKGROUND))
            self.outLines.append("  {} {}\n\n".format(g.GIVEN, self.tree_to_str(nlBatch.context)))

        for scenario in nlBatch.scenarios:
            self.outLines.append("{}: {}\n".format(g.SCENARIO, scenario.name))

            for tree in scenario.givenStatements:
                self.outLines.append("  {} {}\n".format(g.GIVEN, self.tree_to_str(tree)))
            for tree in scenario.whenStatements:
                self.outLines.append("  {} {}\n".format(g.WHEN, self.tree_to_str(tree)))
            for tree in scenario.thenStatements:
                self.outLines.append("  {} {}\n".format(g.THEN, self.tree_to_str(tree)))

    def write(self, outputFilePath: str):
       with open(outputFilePath, "w") as out:
           out.writelines(self.outLines)

    def tree_to_str(self, tree: StatementTree):
        tp = TreeProcessor(self.language)

        # loads yaml file with templates
        tp.load_templ_dic('utils/templates_dic.yaml')
        tree.root = tp.process_tree(tree.root, TreeProcessor.Direction.NL_TO_SIGNAL)

        return self.node_to_str(tree.root)

    def node_to_str(self, node: StatementTreeNode):
        if type(node) == StatementTreeBinaryOperatorNode:
            return self.operator_to_str(node)
        elif type(node) == StatementTreeMergedNode:
            return self.merged_operator_to_str(node)
        elif type(node) == StatementTreeOperandNode:
            return self.operand_to_str(node)

    def merged_operator_to_str(self, node: StatementTreeMergedNode):
        return " && ".join([self.operand_to_str(subnode) for subnode in node.subnodes])

    def operator_to_str(self, node: StatementTreeBinaryOperatorNode):
        if node.kind == c.AND:
            operator = "&&"
        elif node.kind == c.OR:
            operator = "||"
        else:
            raise NotImplementedError("Unrecognized operator {}".format(node.kind))

        return ("({}) {} ({})".format(
            self.node_to_str(node.lchild),
            operator,
            self.node_to_str(node.rchild)))

    def operand_to_str(self, node: StatementTreeOperandNode):
        kind = node.kind

        if kind == c.EQUALITY:
            return "{} == {}".format(
                node.data.variable,
                node.data.value
            )
        elif kind == c.INEQUALITY:
            return  "{} != {}".format(
                node.data.variable,
                node.data.value
            )
        elif kind == c.BOOL:
            return "{}{}".format(
                "" if node.data.value else "! ",
                node.data.variable,
            )
        elif kind in [c.EDGE,
                      c.FORCE,
                      c.UNFORCE]:
            return "{}({}, {})".format(
                node.kind,
                node.data.variable,
                node.data.value
            )
        elif kind == "ACTION":
            # TODO recognize action by NL2Temp
            return "{}({}, {})".format(
                node.kind,
                node.data.variable,
                node.data.value
            )
        else:
            warnings.warn("Node kind {} not recognized!".format(kind))


from gherkan.decoder.NLParser import NLParser
from gherkan.encoder.SignalFileWriter import SignalFileWriter
import importlib
import os

import gherkan.utils.constants as c

# Path to the text file with the signal batch
input_path = os.path.join(c.DATA_DIR, 'input')
output_path = os.path.join(c.DATA_DIR, 'output')
input_file = os.path.join(input_path, 'R1NormalSequenceE.feature')
output_file = os.path.join(output_path, 'R1NormalSequence_signals.feature')

nlFile = NLParser()
# Parse the entire file
nlBatch = nlFile.parseFile(input_file)

signalFileWriter = SignalFileWriter(nlBatch)
signalFileWriter.write(output_file)




import requests
import unittest


class TestNLText(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"
        self.request = {
            "feature" : "robot R1",
            "feature_desc" : "lorem ipsum",
            "background" : "given Line is On",
            "text_raw" : "scenario As soon as shuttle XY is in the station ZX, then robot r1 picks "
                        "up cube one. scenario when robot r3 picks up cube one, robot r2 assembles the product.",
            "language" : "en"
        }

    def test_post_nltext(self):
        requests.get(self.url.format("reset"))
        correctResponse = {
            "info": "Processing done",
            "lines": [],
            "error_lines": [],
            "error_hints": [],
            "errors": False
        }

        response = requests.post(self.url.format("nltext"), data=self.request).json()
        self.assertDictContainsSubset(correctResponse, response)

    # @unittest.skip("TODO: skipped for debugging")
    def test_post_and_get_signal(self):
        requests.get(self.url.format("reset"))
        correctResponse = """# language: en

Feature: robot R1
    lorem ipsum
Background:
  Given LineOn == [empty]

Scenario: As soon as shuttle XY is in the station ZX, then robot r1 picks up cube one. 
  When shuttleXYAtStationZX == [empty]
  Then robotR1ProgramNumber == 1
Scenario: when robot r3 picks up cube one, robot r2 assembles the product.
  When robotR3ProgramNumber == 1
  Then robotR2ProgramNumber == 1
"""

        response = requests.post(self.url.format("nltext"), data=self.request).json()

        response = requests.get(self.url.format("signals"))

        self.assertMultiLineEqual(correctResponse, response.content.decode())




from gherkan.containers.signal_conditional_statement import SignalCondition


class SignalStatementTreeNode():

    def __init__(self, kind=None, data=None):
        self.__lchild = None
        self.__rchild = None
        self.__kind = kind
        self.__data = data
        self.__parent = None

    def setParent(self, parent):
        self.__parent = parent

    @property
    def kind(self):
        return self.__kind

    @property
    def parent(self):
        return self.__parent

    @property
    def data(self):
        return self.__data

    @property
    def lchild(self):
        return self.__lchild

    @property
    def rchild(self):
        return self.__rchild

    @property
    def getIncompleteParent(self):
        raise NotImplementedError()


class SignalStatementTreeOperatorNode(SignalStatementTreeNode):
    def __init__(self, kind):
        super().__init__(kind)
        self.__lchild = None  # type: SignalStatementTreeNode
        self.__rchild = None  # type: SignalStatementTreeNode
        self.templateID = None
        self.variableNLP = None
        self.variableNL_full = None
        self.variableNLPCZ = None
        self.variableNL_full_cs = None

    @property
    def lchild(self):
        return self.__lchild

    @property
    def rchild(self):
        return self.__rchild

    @property
    def children(self):
        return [self.lchild, self.rchild]

    @property
    def full(self):
        return (self.lchild is not None) and (self.rchild is not None)

    def getIncompleteParent(self):
        if self.full:
            if self.parent:
                return self.parent.getIncompleteParent()
            else:
                return None
        else:
            return self

    def attachChild(self, child: SignalStatementTreeNode):
        """
        Returns True if the node is "full" (i.e. both children are assigned)
        """
        child.setParent(self)
        if self.lchild:
            self.__rchild = child
            return True
        else:
            self.__lchild = child
            return False

    def __str__(self):
        return 'Operator(type: {})'.format(self.kind.upper())


class SignalStatementTreeOperandNode(SignalStatementTreeNode):

    def __init__(self, data: SignalCondition):
        super().__init__(data.kind, data)

    def getIncompleteParent(self):
        self.parent.getIncompleteParent()

    def __str__(self):
        return 'Operand(type: {}, variable: {}, value: {}, variableNLP: {}, variableNL_full: {})'.format(
            self.kind.upper() if self.kind else 'None', self.data.variable, self.data.value, self.data.variableNLP,
            self.data.variableNL_full)

    @property
    def variable(self):
        return self.data.variable

    @property
    def kind(self):
        return self.data.kind

    @property
    def variableNLP(self):
        return self.data.variableNLP

    @property
    def variableNL_full(self):
        return self.data.variableNL_full

    @property
    def value(self):
        return self.data.value



import re

from gherkan.containers.SignalStatementTreeNode import SignalStatementTreeNode, SignalStatementTreeOperatorNode


class SignalPhrase:
    def __init__(self):
        self.language = None
        self.tempName = None
        self.tempStr = None
        self.tempToNL = None
        self.inclinations = None
        self.vars = {}
        self.valTypes = {}
        self.niceStrCZ = None
        self.niceStrEN = None


class SignalStatementTree():
    dash, line, left, right = '\u2500', '\u2502', '\u251c\u2500\u2500 ', '\u2514\u2500\u2500 '

    def __init__(self, root: SignalStatementTreeNode, string: str):
        self.__root = root
        self.__string = string

    @property
    def root(self):
        return self.__root

    @property
    def string(self):
        return self.__string

    def __printNode(self, node: SignalStatementTreeNode, leftChild, levelsOpened):
        string = ''.join([SignalStatementTree.line + "\t" if lo else "\t" for lo in levelsOpened[:-1]])
        if leftChild:
            if node:
                string += SignalStatementTree.left + str(str(node) if bool(node) else "ERROR!!!") + '\n'
        else:
            if node:
                string += SignalStatementTree.right + str(str(node) if bool(node) else "ERROR!!!") + '\n'

        if type(node) is SignalStatementTreeOperatorNode:
            string += self.__printNode(node.lchild, True, levelsOpened + [True])
            string += self.__printNode(node.rchild, False, levelsOpened + [False])
        return string

    def __str__(self):
        levelsOpened = [False]

        node = self.root
        string = str(node) + '\n'
        if type(node) is SignalStatementTreeOperatorNode:
            string += self.__printNode(node.lchild, True, levelsOpened + [True])
            string += self.__printNode(node.rchild, False, levelsOpened + [False])
        return string






import re

import pkg_resources
import yaml
import warnings
from gherkan.containers.SignalPhrase import SignalPhrase
from gherkan.containers.StatementTreeNode import StatementTreeBinaryOperatorNode, StatementTreeOperandNode, \
    StatementTreeNode, StatementTreeMergedNode

from gherkan.encoder.NL2Temp import NLTempMatcher

from enum import Enum
from pprint import pprint as pp
from gherkan.processing.Temp2NL import FillActions
from gherkan.utils import constants as c


class TreeProcessor:
    class Direction(Enum):
        SIGNAL_TO_NL = 0
        NL_TO_SIGNAL = 1

    def __init__(self, language):
        self.templ_dic = None
        self.language = language


    def process_tree(self, node: StatementTreeNode, dir: Direction):
        if dir not in [TreeProcessor.Direction.SIGNAL_TO_NL,
                       TreeProcessor.Direction.NL_TO_SIGNAL]:
            raise NotImplementedError("Direction {} unknown".format(dir))

        self.dir = dir

        return self.process_tree_recursive(node)


    def process_tree_recursive(self, node: StatementTreeNode):
        if type(node) is StatementTreeBinaryOperatorNode:

            if self.dir == TreeProcessor.Direction.SIGNAL_TO_NL and node.kind == c.AND:
                # try to match compound template which can be rooted in the AND node
                res = self.match_compound_signal_templates(node)

                # in case the compound template has been matched, replace the AND node
                # at the parent by the new merged node
                if res:
                    template, matches, merged_node = res
                    node = self.process_operand(merged_node, matches, template)
                    return merged_node

            node.lchild = self.process_tree_recursive(node.lchild)
            node.rchild = self.process_tree_recursive(node.rchild)

        elif type(node) is StatementTreeOperandNode:

            if self.dir == TreeProcessor.Direction.NL_TO_SIGNAL:
                # normalize NL phrase before template matching
                ntm = NLTempMatcher(lang=self.language)
                res = ntm.get_NLPhrase(node.data.variableNL)
                node.data.variableNL = res["tempToNL"]

                match, template = self.match_simple_templates(node)

                if match is None:
                    warnings.warn("No match found for phrase: " + node.data.variableNL)
                    return

                node = self.process_operand(node, [match], template, negate=res["negate"])

            elif self.dir == TreeProcessor.Direction.SIGNAL_TO_NL:
                match, template = self.match_simple_templates(node)

                if match is None:
                    warnings.warn("No match found for phrase: " + node.data.string)
                    return

                node = self.process_operand(node, [match], template)

        else:
            raise NotImplementedError("Unrecognized type of node: {}".format(type(node)))

        # returned node should be reassigned in case the node changed
        return node


    def process_operand(self, node: StatementTreeNode, matches: list, template, negate: bool = False):
        sp = SignalPhrase()
        sp.tempName = template["Name"]
        sp.tempStr = template["SignalTemplate"]
        sp.tempToNL = template["NLTemplateEn"] if self.language == c.LANG_EN else template["NLTemplateCz"]
        sp.inclinations = template["Inclinations"]
        sp.language = self.language

        for match in matches:
            for key, val in match.groupdict().items():
                sp.vars[key] = val

        if self.dir == TreeProcessor.Direction.SIGNAL_TO_NL:
            if type(node) == StatementTreeOperandNode:
                sp.values[node.data.valueId] = node.data.value

            elif type(node) == StatementTreeMergedNode:
                for subnode in node.subnodes:
                    sp.values[subnode.data.valueId] = subnode.data.value

            self.translate_to_nl(node, sp)
        elif self.dir == TreeProcessor.Direction.NL_TO_SIGNAL:
            node = self.translate_to_signal(node, sp, template, negate=negate)

        return node



    def translate_to_signal(self, node: StatementTreeOperandNode, sp: SignalPhrase, template: dict, negate: bool):
        signal_templates = template["SignalTemplate"]

        if len(signal_templates) > 1:
            mergedNode = StatementTreeMergedNode(parent=node.parent)

            for signal_template in signal_templates:
                subnode = StatementTreeOperandNode(kind=None) # kind is filled by translate_node_to_signal()
                subnode = self.translate_node_to_signal(subnode, sp, signal_template, negate=negate)
                mergedNode.subnodes.append(subnode)

                if negate:
                    # AND -> OR
                    mergedNode.negate()

            return mergedNode

        else:
            return self.translate_node_to_signal(node, sp, template["SignalTemplate"][0], negate=negate)


    def translate_node_to_signal(self, node: StatementTreeOperandNode, sp: SignalPhrase, signal_template: dict, negate: bool):
        variable = signal_template["Template"]

        # replace all named patterns by their value
        for group, value in sp.vars.items():
            if value is not None:
                strToReplace = re.compile(r"\(\?P\<{}\>[^\)]+\)".format(group))
                variable = re.sub(strToReplace, value, variable)

        # the result should be a variable name without any whitespaces
        node.kind = signal_template["Type"]
        node.data.variable = re.sub(r'\s+', '', variable)

        # fill in the value
        valueId = signal_template["ValueId"]

        if valueId in sp.vars:
            # should work for every template where value is stated explicitly in the NL phrase
            node.data.value = sp.vars[valueId]
        elif node.kind == c.BOOL:
            # default value for bool
            node.data.value = True
        else:
            # TODO reality check... this should not occur in templates in practice
            # default value for everything else
            node.data.value = 1

        if negate:
            node.negate()

        return node


    def translate_to_nl(self, node: StatementTreeNode, signalPhrase: SignalPhrase):
        # call Temp2NL
        fill_action = FillActions()
        fill_action.ParseMainAction(data=signalPhrase)

        node.data.variableNL = signalPhrase.niceStr

        # #TODO only debug
        # if type(node) == StatementTreeOperandNode:
        #     node.data.variableNL = "NL[{}={}]".format(node.data.variable, node.data.value)
        # elif type(node) == StatementTreeMergedNode:
        #     node.data.variableNL = "NL[merged_node]"



    def match_compound_signal_templates(self, node: StatementTreeBinaryOperatorNode):
        # only for direction TreeProcessor.Direction.SIGNAL_TO_NL
        operands = self.gather_conjunction_operands(node)

        for index, template in self.templ_dic.items():
            if not operands or len(template["SignalTemplate"]) != len(operands):
                # number of operands do not equal number of template strings,
                # cannot match the compound template
                continue

            matched_str_idxs = []
            matches = []

            for operand in operands:
                i, match = self.match_signal_template(operand, template)

                if match:
                    matches.append(match)
                    # save the index of the matched template string
                    matched_str_idxs.append(i)

            # check if matched templates strings are unique (none was matched twice)
            # and complete (all were matched)
            if len(set(matched_str_idxs)) == len(operands):
                merged_node = StatementTreeMergedNode()
                merged_node.subnodes = operands

                return template, matches, merged_node

        return None


    def gather_conjunction_operands(self, node):
        # gathers a list of operand nodes in a subtree connected via AND operator nodes
        if node.kind == c.AND:
            left_operands = self.gather_conjunction_operands(node.lchild)
            right_operands = self.gather_conjunction_operands(node.rchild)

            if left_operands and right_operands:
                # concatenate lists from both branches
                return left_operands + right_operands

            # if one of the branches did not succeed, the whole operation failed

        elif type(node) is StatementTreeOperandNode:
            # we encountered an operand, return it as an element of a list
            return [node]

        # signalizes the operation was invalid, subtree is not a conjunction of operands
        return None




    def match_simple_templates(self, node: StatementTreeOperandNode):
        # finds a matching template in the YAML file (works for both directions)
        match = None

        for template in self.templ_dic.values():
            if self.dir == TreeProcessor.Direction.NL_TO_SIGNAL:
                _, match = self.match_nl_template(node, template)
            elif self.dir == TreeProcessor.Direction.SIGNAL_TO_NL:

                if len(template["SignalTemplate"]) > 1:
                    # here we match only simple templates
                    continue

                _, match = self.match_signal_template(node, template)

            if match:
                return match, template

        return None, None



    def match_signal_template(self, operand: StatementTreeOperandNode, template): #TODO template type
        templates_str = template["SignalTemplate"]

        # at least one of the SignalTemplates has to match operand variable
        for i, template_str in enumerate(templates_str):
            template_type = template_str["Type"]

            # kind of operand has to match the template type
            if template_type == operand.kind \
                    or template_type == "ACTION" and operand.kind in [c.UNFORCE, c.FORCE, c.EDGE]:  # TODO only testing, groups should be defined elsewhere

                expToMatch = re.compile(r"^{}$".format(template_str["Template"].strip()))
                match = expToMatch.search(operand.data.variable)

                if match:
                    operand.data.valueId = template_str["ValueId"]
                    return i, match

        # no match found
        return None, None

    def match_nl_template(self, operand: StatementTreeOperandNode, template): #TODO template type
        templates_str = template["NLTemplateEn"
                            if self.language == c.LANG_EN
                            else "NLTemplateCz"]


        for i, template_str in enumerate(templates_str):
            expToMatch = re.compile(r"^\s*{}\s*$".format(template_str))
            match = expToMatch.search(operand.data.variableNL)

            if match:
                return i, match

        return None, None


    def load_templ_dic(self, file):
        # load templates to be matched
        with pkg_resources.resource_stream("gherkan", file) as stream:
            try:
                self.templ_dic = yaml.load(stream)
            except yaml.YAMLError as exc:
                print(exc)


"""
This class aligns natural language phrases into their most default forms and matches them
with niceStr parameters located in templates_dic.yaml. Returns an dict with selected template and
filled as many fields from the dic as possible.
"""
import re
from gherkan.utils.word_morpho import Finetuner
from pattern.en import parse, parsetree, lemma
import majka
import pkg_resources, json



class NLTempMatcher():
    def __init__(self, lang):
        self.finetune = Finetuner()
        self.NLPhrase = None
        self.lang = lang
        self.program_dic = self.load_program_dic()

    def get_NLPhrase(self, phrase):
        negative = self.get_negative(phrase)
        force = self.get_force(phrase)
        clean_phrase = self.get_cleanphr(phrase, negative, force)
        clean_phrase = self.lemmatize(clean_phrase)
        pn = self.get_pn(clean_phrase)
        if pn:
            clean_phrase = pn
        clean_phrase = self.finetune.strip_extra_spaces(clean_phrase)
        self.NLPhrase = {"tempToNL": clean_phrase,
                         "negate": negative,
                         "force": force
                         }
        return self.NLPhrase

    def get_pn(self, phrase):
        tree = parsetree(phrase, tokenize=True,
                         tags=True, chunks=True, relations=True)
        for sentence in tree:
            try:
             actor = str(sentence.subjects[0]).lower()
            except:
                return False
        action_lemma = self.lemmatize("".join(phrase.rsplit(actor)))
        if actor in self.program_dic.keys():
            for key in self.program_dic[actor]:
                if self.finetune.strip_extra_spaces(action_lemma) in self.program_dic[actor][key]:
                    phrase = " ".join([actor, "program number be", key])
                    return phrase
        return False

    # def get_edge(self, phrase): # TODO
    #     edge = phrase
    #     return 1

    def get_force(self, phrase):
        if self.lang == 'en':
           force = re.findall(r'\b(force)\b', phrase, re.IGNORECASE)
        elif self.lang == 'cs':
            no_diacritics = self.finetune.remove_diacritics(phrase)
            force = re.findall(r'\b(vynut)\b', no_diacritics, re.IGNORECASE)
        if not self.empty_tree(force):
            force = True
        else:
            force = False
        return force

    def get_negative(self, phrase):
        verbs = self.finetune.find_verb(self.lang, phrase)
        if self.lang == "en":
            if verbs:
                for verb in verbs:
                    negative = re.findall(r'\b(not)\b\s+\b({})\b|(\b({})\b)\s\b(not)\b'.format(verb, verb), phrase)
                    if not self.empty_tree(negative):
                        negative = True
                    else:
                        negative = False
            else:
                negative = False
        elif self.lang == "cs":
            if verbs:
                for verb in verbs:
                    pass

        return negative

    def get_cleanphr(self, phrase, negative=None, force=None):
        # clean the phrase from negatives and force
        if self.lang == "en":
            if negative == True:
              phrase = re.sub(r'\b(not)\b', '', phrase)
            if force == True:
              phrase = re.sub(r'\b(force)\b', '', phrase)
        elif self.lang == "cs":
          pass

        # remove excess verb words
        for sent in parsetree(phrase, tags=True, chunks=True, relations=True, lemmas=True):
            del_words = ["do", "will", "can", "must", "have to"]
            if sent.verbs:
                for i in sent.verbs[0].string.split():
                   if lemma(i) in del_words:
                    action_cl = re.sub(r'^\b({})\b\s'.format(i), "", sent.verbs[0].string)
                    phrase = self.finetune.strip_extra_spaces(phrase)
                    phrase = re.sub(sent.verbs[0].string, action_cl, phrase)

        # print(parse(clean_phrase, relations=True, lemmata=True))
        return phrase

    def empty_tree(self, input_list):
        """Recursively iterate through values in nested lists."""
        for item in input_list:
            if not isinstance(item, list) or not self.empty_tree(item):
                return False
        return True

    def lemmatize(self, text):
        lemmas = ""
        skip_list = [",", ".", ";"]
        if self.lang == "cs":
            self.morph_wlt.flags |= majka.ADD_DIACRITICS  # find word forms with diacritics
            self.morph_wlt.tags = False  # return just the lemma, do not process the tags
            self.morph_wlt.first_only = True  # return only the first entry
            for word in text.split():
               if word in skip_list:
                   pass
               else:
                   output = self.morph_wlt.find(word)
                   lemma_det = output[0]['lemma']
                   lemmas += " "
                   lemmas += lemma_det

        elif self.lang == "en":
            # this version keeps capitals
            tree = parsetree(text, tokenize=True)

            for sentence in tree:
                for word in sentence:
                    if word.string in skip_list:
                        pass
                    else:
                        lemmas += " "
                    lemmas += lemma(word.string)

        return lemmas

    def load_program_dic(self):
        # load program json
        if self.lang == "en":
            file = "utils/RobotPrograms_en.json"
        elif self.lang == "cs":
            file = "utils/RobotPrograms_cz.json"
        with pkg_resources.resource_stream("gherkan", file) as stream:
            try:
                program_dic = json.load(stream)
            except:
                print("Could not load {}".format(file))
        return program_dic


if __name__ == "__main__":
   match_temp = NLTempMatcher()



from gherkan.decoder.NLParser import NLParser
from gherkan.encoder.RawNLParser import RawNLParser
from gherkan.encoder.SignalFileWriter import SignalFileWriter

def process(base_path: str, request: dict):
    """
    Processes raw text and saves two files - intermediate NL file (base_path.feature) and
    resulting signal file (base_path_signals.feature).

    Parameters
    ----------
    base_path : str
        path/to/files/file_basename

    request : dict
        dictionary with raw text
    """

    # create paths to files based on base_path
    nl_file_path = base_path + ".feature"
    signal_file_path = base_path + "_signals.feature"

    # parse the raw text
    parser = RawNLParser()
    parser.parse(request["text_raw"])
    request["background"] = parser.process_sentence(request["background"])
    parser.generate_program_dict()
    raw_text_lines = parser.get_text_lines()

    # create intermediate representation - text lines of the NL file
    lines = [
        "# language: {}".format(request["language"]),
        "Feature: {}".format(request["feature"]),
        "  {}".format(request["feature_desc"]),
        "Background:",
        "  {}".format(request["background"])
    ]
    lines += raw_text_lines

    # save the intermediate NL file
    with open(nl_file_path, "wt") as out:
        out.write("\n".join(lines))

    print("Saved NL file as {}".format(nl_file_path))

    # parse the NL text lines
    nl_parser = NLParser()
    nl_batch = nl_parser.parse(lines)

    # write the signal file
    signalFileWriter = SignalFileWriter(nl_batch)
    signalFileWriter.write(signal_file_path)

    print("Saved signals file as {}".format(signal_file_path))



from flask import jsonify, request
from flask_restful import Resource, reqparse


class SignalMap(Resource):

    def get(self):
        """
        response = {
            "internalSignalName1": "externalSignalName1",
            "internalSignalName2": "externalSignalName2"
        }
        """
        pass

    def post(self):
        """
        request = {
            "internalSignalName1": "externalSignalName1",
            "internalSignalName2": "externalSignalName2"
        }
        """
        pass




import speech_recognition as sr
import argparse
import os

"""
Speech-To-Text transcription using Google API.
Currently, SpeechRecognition supports the following file formats:

WAV: must be in PCM/LPCM format
AIFF
AIFF-C
FLAC: must be native FLAC format; OGG-FLAC is not supported

Example od usage:
python3 speech_to_text,py --audio_file ./myfile.wav --lang cs-CZ

"""


def process(audio_file_path: str, lang: str):
    rec = sr.Recognizer()
    audio_file = sr.AudioFile(audio_file_path)

    transcriptPath = '.'.join(audio_file_path[:audio_file_path.find(".")], ".txt")

    with audio_file as source:
        print("Processing audio...")
        audio = rec.record(source)
        recognized = rec.recognize_google(audio, language=lang)
        print("Recognized text: {}".format(recognized))

    with open(transcriptPath, "w") as tf:
        tf.writelines(recognized)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--audio_file", help="Path to input audio file")
    parser.add_argument("--lang", help="en-US/en-UK/cs-CZ")
    args = parser.parse_args()

    rec = sr.Recognizer()
    audio_file = sr.AudioFile(args.audio_file)

    with audio_file as source:
        print("Processing audio...")
        audio = rec.record(source)
        recognized = rec.recognize_google(audio, language=args.lang)
        print("Recognized text: {}".format(recognized))


# coding: utf-8
from googletrans import Translator
import majka

class AdjustWords():

    def lemmatize(self, word):
        """  Returns lemma of given word form / Vrátí základní podobu slova (např. infinitiv)"""

        morph = majka.Majka('./majka.w-lt')
        morph.flags |= majka.ADD_DIACRITICS  # find word forms with diacritics
        morph.tags = False  # return just the lemma, do not process the tags
        morph.first_only = True  # return only the first entry
        output = morph.find(word)
        lemma = output[0]['lemma']

        return lemma


    def change_case(self, word, case, plural = False):
        """  Returns provided noun in given grammatical case / Vrátí podstatné jméno vyskloňované v daném pádu """
        morph = majka.Majka('./majka.l-wt')
        morph.flags |= majka.ADD_DIACRITICS
        word_form = []
        analysis = morph.find(word)

        if plural == False:
            for n in range(len(analysis)):
                if 'singular' in analysis[n]['tags']:
                    declen = [analysis[n]['lemma'],analysis[n]['tags']['case']]
                    if case in declen:
                       word_form.append(declen[0])
                else:
                    pass
        else:
            for n in range(len(analysis)):
                if 'plural' in analysis[n]['tags']:
                    declen = [analysis[n]['lemma'], analysis[n]['tags']['case']]
                    if case in declen:
                        word_form.append(declen[0])
            else:
                pass

        return word_form[0]

    def get_cases(self, word):
        """ Returns all possible cases of the word, in case of verb returns all info / Vrátí možné pády daného slova, u sloves vrátí všechno info"""
        morph = majka.Majka('./majka.w-lt')
        morph.flags |= majka.ADD_DIACRITICS  # find word forms with diacritics
        output = morph.find(word)

        cases = []
        if "substantive" in output[0]['tags']['pos'] or "adjective" in output[0]['tags']['pos']:
            for n in range(len(output)):
                 case = output[n]['tags']['case']
                 if "singular" in output[n]['tags']:
                  number = "singular"
                 else:
                  number = "plural"
                 cases.append([case, number])

        elif "verb" in output[0]['tags']['pos']:
            cases = output

        # for x in cases:
        #     if x < final_list:
        #         pass
        #     else:
        #         final_list.append(x)

        return cases

    def translate(self, phrase, lang = "cz-en"):
        translator = Translator()
        if lang == "cz-en":
          result = translator.translate(phrase, src="cs", dest='en')
        elif lang == "en-cz":
          result = translator.translate(phrase, src="en", dest='cs')

        translation = result.text
        return translation



if __name__ == "__main__":

 entry = "put on the line"
 adjust = AdjustWords()
 word = adjust.translate(entry, "en-cz")
print(word)



# coding: utf-8
from gherkan.processing import Temp2NL_nomajka
fill_action = Temp2NL_nomajka.FillActions()

VarsVals = ['robotR1', '4']
Type = "equality"
Values = "4"
PhraseStr = "ProgramNumber"

TempData = [VarsVals, Type, Values, PhraseStr]
output = fill_action.ParseMainAction(data=TempData)
print(output)




import unittest
import gherkan.utils.constants as c
import os
"""
This code will gather and run all the tests in this directory. Files with tests must start with the "test_" prefix.
"""

unittest.TestCase.host = "localhost"
unittest.TestCase.port = 5000

loader = unittest.TestLoader()
testSuite = loader.discover(os.path.dirname(os.path.abspath(__file__)))

testRunner = unittest.TextTestRunner(verbosity=2)
testRunner.run(testSuite)



import yaml
import re

with open('templates_dic.yaml','r') as stream:
    try:
        temp_dic = yaml.load(stream)
    except yaml.YAMLError as exc:
        print(exc)

#robotXProgram = re.compile(r"(?=.*OR)(?=.*AND)(?=.*robot(?P<robotName>\w+)programnumber)", re.IGNORECASE)#(?=(?P<robotName>\w+)programnumber)\\1+(programended)
#match = robotXProgram.search("AND(OR(robotN3ProgramNumber,robotN4ProgramNumber),programended)")  # parse the string using the regex
#print(match)
#print("Robot name was '{}'".format(match.group("robotName")))


stringToEval = "HalloOnBox"
for index, tmplD in temp_dic.items():
    print(tmplD["TempStr"][0])
    strAll = ''.join(tmplD["TempStr"])
    print(strAll)
    expToMatch = re.compile(r""+tmplD["TempStr"][0], re.IGNORECASE)
    #expToMatch = re.compile(r""+"(?P<subject>\\w+)On(?P<object>\\w+)", re.IGNORECASE)
    match = expToMatch.search(stringToEval)
    if match:
        print(match[0])
        print(match.group)
        print("Robot name was '{}'".format(match.group("subject")))
# dic = {1:{"brand":"ford","type":"mustang","year":"2018"},2:{"brand":"ford1","type":"mustang1","year":"20181"}}
# for index, d in dic.items():
#     print(d["brand"])
#     if "brand" in d.keys():
#         print("hallo maus")
